###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.datetime.date import date, Date
from everysk.core.datetime.datetime import datetime, DateTime
from everysk.core.exceptions import DateError
from everysk.core.unittests import TestCase


####################################################################################
#                                    Date tests                                    #
####################################################################################
class DateTestCase(TestCase):

    def test_ensure_date(self):
        ret = Date.ensure(date(2023, 1, 1))
        self.assertEqual(ret.__class__, Date)
        self.assertEqual(ret, Date(2023, 1, 1))

        ret = Date.ensure(Date(2023, 1, 1))
        self.assertEqual(ret.__class__, Date)
        self.assertEqual(ret, Date(2023, 1, 1))

    def test_ensure_datetime(self):
        ret = Date.ensure(datetime(2023, 1, 1, 12, 13, 14, 1516)) # pylint: disable=not-callable
        self.assertEqual(ret.__class__, Date)
        self.assertEqual(ret, Date(2023, 1, 1))

        ret = Date.ensure(DateTime(2023, 1, 1, 12, 13, 14, 1516))
        self.assertEqual(ret.__class__, Date)
        self.assertEqual(ret, Date(2023, 1, 1))

    def test_ensure_date_invalid(self):
        with self.assertRaisesRegex(ValueError, "Invalid instantiation of class 'Date' from 'str'"):
            Date.ensure('20210101')

    def test_date_fromisoformat(self):
        result = Date.fromisoformat('20230101')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 1, 1))

        result = Date.fromisoformat('2023-01-01')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 1, 1))

    def test_date_fromordinal(self):
        result = Date.fromordinal(1)
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(1, 1, 1))

        result = Date.fromordinal(365, start_date=Date(1, 1, 1))
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(1, 12, 31))

        result = Date.fromordinal(42000, start_date=Date(2020, 1, 1))
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2134, 12, 28))

    def test_date_strptime(self):
        result = Date.strptime('20230731')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 7, 31))

        result = Date.strptime('20230731', format='%Y%m%d')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 7, 31))

        result = Date.strptime('2023-07-31', format='%Y-%m-%d')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 7, 31))

    def test_date_strptime_or_null(self):
        result = Date.strptime_or_null('20230731')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 7, 31))

        result = Date.strptime_or_null('20230731', format='%Y%m%d')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 7, 31))

        result = Date.strptime_or_null('2023-07-31', format='%Y-%m-%d')
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, Date(2023, 7, 31))

        self.assertIsNone(Date.strptime_or_null('2023-07-31'))
        self.assertIsNone(Date.strptime_or_null('2023-07-31', format='%Y%m%d'))
        self.assertIsNone(Date.strptime_or_null(None))
        self.assertIsNone(Date.strptime_or_null(None, format='%Y-%m-%d'))

    def test_date_is_string_date(self):
        self.assertTrue(Date.is_string_date('20230731'))
        self.assertTrue(Date.is_string_date('20230731', format='%Y%m%d'))
        self.assertTrue(Date.is_string_date('2023-07-31', format='%Y-%m-%d'))

        self.assertFalse(Date.is_string_date('2023-07-31'))
        self.assertFalse(Date.is_string_date('2023-07-31', format='%Y%m%d'))
        self.assertFalse(Date.is_string_date(None))
        self.assertFalse(Date.is_string_date(None, format='%Y-%m-%d'))

    def test_date_strftime(self):
        self.assertEqual(Date(2023, 7, 31).strftime(), '20230731')
        self.assertEqual(Date(2023, 7, 31).strftime('%Y%m%d'), date(2023, 7, 31).strftime('%Y%m%d'))
        self.assertEqual(Date(2023, 7, 31).strftime(format='%Y%m%d'), '20230731')
        self.assertEqual(Date(2023, 7, 31).strftime(format='%Y-%m-%d'), '2023-07-31')

    def test_date_strftime_or_null(self):
        self.assertEqual(Date.strftime_or_null(Date(2023, 7, 31)), '20230731')
        self.assertEqual(Date.strftime_or_null(Date(2023, 7, 31), format='%Y%m%d'), '20230731')
        self.assertEqual(Date.strftime_or_null(Date(2023, 7, 31), format='%Y-%m-%d'), '2023-07-31')

        self.assertEqual(Date.strftime_or_null(date(2023, 7, 31)), '20230731')
        self.assertEqual(Date.strftime_or_null(date(2023, 7, 31), format='%Y%m%d'), '20230731')
        self.assertEqual(Date.strftime_or_null(date(2023, 7, 31), format='%Y-%m-%d'), '2023-07-31')

        self.assertEqual(Date.strftime_or_null(DateTime(2023, 7, 31, 12, 13, 14)), '20230731')
        self.assertEqual(Date.strftime_or_null(DateTime(2023, 7, 31, 12, 13, 14), format='%Y%m%d'), '20230731')
        self.assertEqual(Date.strftime_or_null(DateTime(2023, 7, 31, 12, 13, 14), format='%Y-%m-%d'), '2023-07-31')

        self.assertEqual(Date.strftime_or_null(datetime(2023, 7, 31, 12, 13, 14)), '20230731') # pylint: disable=not-callable
        self.assertEqual(Date.strftime_or_null(datetime(2023, 7, 31, 12, 13, 14), format='%Y%m%d'), '20230731') # pylint: disable=not-callable
        self.assertEqual(Date.strftime_or_null(datetime(2023, 7, 31, 12, 13, 14), format='%Y-%m-%d'), '2023-07-31') # pylint: disable=not-callable

        self.assertIsNone(Date.strftime_or_null('2023-07-31'))
        self.assertIsNone(Date.strftime_or_null('2023-07-31', format='%Y%m%d'))
        self.assertIsNone(Date.strftime_or_null(None))
        self.assertIsNone(Date.strftime_or_null(None, format='%Y-%m-%d'))

    def test_date_is_date(self):
        self.assertTrue(Date.is_date(date(2023, 7, 31)))
        self.assertTrue(Date.is_date(Date(2023, 7, 31)))

        self.assertFalse(Date.is_date(DateTime(2023, 7, 31)))
        self.assertFalse(Date.is_date(DateTime(2023, 7, 31)))
        self.assertFalse(Date.is_date(None))

    def test_date_today(self):
        result = Date.today()
        self.assertEqual(result.__class__, Date)
        self.assertEqual(result, date.today())

    def test_date_is_today(self):
        self.assertTrue(Date.is_today(Date.today()))
        self.assertFalse(Date.is_today(Date(2023, 1, 1)))

    def test_is_date(self):
        self.assertTrue(Date.is_date(Date.today()))
        self.assertTrue(Date.is_date(Date(2023, 1, 1)))
        self.assertTrue(Date.is_date(date(2023, 1, 1)))
        self.assertFalse(Date.is_date(DateTime.now()))
        self.assertFalse(Date.is_date('date'))
        self.assertFalse(Date.is_date(2022))

    def test_date_quarter(self):
        self.assertEqual(Date(2023, 1, 1).quarter, 1)
        self.assertEqual(Date(2023, 3, 31).quarter, 1)
        self.assertEqual(Date(2023, 4, 1).quarter, 2)
        self.assertEqual(Date(2023, 6, 30).quarter, 2)
        self.assertEqual(Date(2023, 7, 1).quarter, 3)
        self.assertEqual(Date(2023, 9, 30).quarter, 3)
        self.assertEqual(Date(2023, 10, 1).quarter, 4)
        self.assertEqual(Date(2023, 12, 31).quarter, 4)

    def test_date_month_name(self):
        self.assertEqual(Date(2023, 1, 1).month_name, 'January')
        self.assertEqual(Date(2023, 2, 1).month_name, 'February')
        self.assertEqual(Date(2023, 3, 1).month_name, 'March')
        self.assertEqual(Date(2023, 4, 1).month_name, 'April')
        self.assertEqual(Date(2023, 5, 1).month_name, 'May')
        self.assertEqual(Date(2023, 6, 1).month_name, 'June')
        self.assertEqual(Date(2023, 7, 1).month_name, 'July')
        self.assertEqual(Date(2023, 8, 1).month_name, 'August')
        self.assertEqual(Date(2023, 9, 1).month_name, 'September')
        self.assertEqual(Date(2023, 10, 1).month_name, 'October')
        self.assertEqual(Date(2023, 11, 1).month_name, 'November')
        self.assertEqual(Date(2023, 12, 1).month_name, 'December')

    def test_date_week_of_year(self):
        self.assertEqual(Date(2023, 1, 1).week_of_year, 1)
        self.assertEqual(Date(2023, 6, 30).week_of_year, 26)
        self.assertEqual(Date(2023, 9, 30).week_of_year, 39)
        self.assertEqual(Date(2023, 12, 31).week_of_year, 53)

    def test_date_week_of_month(self):
        self.assertEqual(Date(2023, 1, 1).week_of_month, 1)
        self.assertEqual(Date(2023, 6, 8).week_of_month, 2)
        self.assertEqual(Date(2023, 9, 15).week_of_month, 3)
        self.assertEqual(Date(2023, 12, 31).week_of_month, 5)

    def test_date_day_name(self):
        self.assertEqual(Date(2023, 1, 1).day_name, 'Sunday')
        self.assertEqual(Date(2023, 1, 2).day_name, 'Monday')
        self.assertEqual(Date(2023, 1, 3).day_name, 'Tuesday')
        self.assertEqual(Date(2023, 1, 4).day_name, 'Wednesday')
        self.assertEqual(Date(2023, 1, 5).day_name, 'Thursday')
        self.assertEqual(Date(2023, 1, 6).day_name, 'Friday')
        self.assertEqual(Date(2023, 1, 7).day_name, 'Saturday')

    def test_date_day_of_year(self):
        self.assertEqual(Date(2023, 1, 1).day_of_year, 1)
        self.assertEqual(Date(2023, 6, 30).day_of_year, 181)
        self.assertEqual(Date(2023, 9, 30).day_of_year, 273)
        self.assertEqual(Date(2023, 12, 31).day_of_year, 365)

    def test_date_toordinal(self):
        self.assertEqual(Date(1, 1, 1).toordinal(), 1)
        self.assertEqual(Date(1, 12, 31).toordinal(start_date=Date(1, 1, 1)), 365)
        self.assertEqual(Date(2134, 12, 28).toordinal(start_date=Date(2020, 1, 1)), 42000)

    def test_date_get_last_day_of_week(self):
        ret = Date(2023, 1, 1).get_last_day_of_week()
        self.assertEqual(ret, Date(2023, 1, 7))
        self.assertEqual(ret.__class__, Date)
        self.assertEqual(Date(2023, 1, 7).get_last_day_of_week(), Date(2023, 1, 7))
        self.assertEqual(Date(2023, 1, 7).get_last_day_of_week(bizdays=True), Date(2023, 1, 6))
        self.assertEqual(Date(2023, 1, 1).get_last_day_of_week(bizdays=True, calendar='ANBIMA'), Date(2023, 1, 6))
        self.assertEqual(Date(2023, 4, 3).get_last_day_of_week(), Date(2023, 4, 8))
        self.assertEqual(Date(2023, 4, 3).get_last_day_of_week(bizdays=True), Date(2023, 4, 7))
        self.assertEqual(Date(2023, 4, 3).get_last_day_of_week(bizdays=True, calendar='ANBIMA'), Date(2023, 4, 6))

    def test_date_get_last_day_of_month(self):
        self.assertEqual(Date(2018, 5, 15).get_last_day_of_month().__class__, Date)
        self.assertEqual(Date(2018, 5, 15).get_last_day_of_month(), Date(2018, 5, 31))
        self.assertEqual(Date(2018, 5, 15).get_last_day_of_month(bizdays=True), Date(2018, 5, 31))
        self.assertEqual(Date(2018, 5, 15).get_last_day_of_month(bizdays=True, calendar='ANBIMA'), Date(2018, 5, 30))

    def test_date_get_last_day_of_quarter(self):
        self.assertEqual(Date(2023, 1, 15).get_last_day_of_quarter().__class__, Date)
        self.assertEqual(Date(2023, 1, 15).get_last_day_of_quarter(), Date(2023, 3, 31))
        self.assertEqual(Date(2024, 5, 31).get_last_day_of_quarter(), Date(2024, 6, 30))
        self.assertEqual(Date(2023, 4, 15).get_last_day_of_quarter(), Date(2023, 6, 30))
        self.assertEqual(Date(2023, 7, 15).get_last_day_of_quarter(bizdays=True), Date(2023, 9, 29))
        self.assertEqual(Date(2023, 11, 15).get_last_day_of_quarter(bizdays=True, calendar='ANBIMA'), Date(2023, 12, 29))

    def test_date_get_last_day_of_year(self):
        self.assertEqual(Date(2023, 1, 15).get_last_day_of_year().__class__, Date)
        self.assertEqual(Date(2023, 1, 15).get_last_day_of_year(), Date(2023, 12, 31))
        self.assertEqual(Date(2023, 7, 15).get_last_day_of_year(bizdays=True), Date(2023, 12, 29))
        self.assertEqual(Date(2023, 11, 15).get_last_day_of_year(bizdays=True, calendar='ANBIMA'), Date(2023, 12, 29))

    def test_date_is_last_day_of_week(self):
        self.assertTrue(Date(2023, 4, 8).is_last_day_of_week())
        self.assertFalse(Date(2023, 4, 7).is_last_day_of_week())
        self.assertFalse(Date(2023, 4, 6).is_last_day_of_week())
        self.assertFalse(Date(2023, 4, 8).is_last_day_of_week(bizdays=True))
        self.assertTrue(Date(2023, 4, 7).is_last_day_of_week(bizdays=True))
        self.assertTrue(Date(2023, 4, 6).is_last_day_of_week(bizdays=True, calendar='ANBIMA'))

    def test_date_is_last_day_of_month(self):
        self.assertTrue(Date(2023, 1, 31).is_last_day_of_month())
        self.assertFalse(Date(2023, 2, 27).is_last_day_of_month())
        self.assertTrue(Date(2023, 12, 31).is_last_day_of_month())
        self.assertTrue(Date(2023, 4, 28).is_last_day_of_month(bizdays=True))
        self.assertFalse(Date(2023, 4, 30).is_last_day_of_month(bizdays=True))
        self.assertFalse(Date(2023, 12, 31).is_last_day_of_month(bizdays=True, calendar='ANBIMA'))

    def test_date_is_last_day_of_quarter(self):
        self.assertFalse(Date(2023, 1, 1).is_last_day_of_quarter())
        self.assertTrue(Date(2023, 3, 31).is_last_day_of_quarter())
        self.assertFalse(Date(2023, 4, 1).is_last_day_of_quarter())
        self.assertTrue(Date(2023, 6, 30).is_last_day_of_quarter())
        self.assertFalse(Date(2023, 7, 1).is_last_day_of_quarter())
        self.assertTrue(Date(2023, 9, 30).is_last_day_of_quarter())
        self.assertFalse(Date(2023, 10, 1).is_last_day_of_quarter())
        self.assertTrue(Date(2023, 12, 31).is_last_day_of_quarter())
        self.assertTrue(Date(2023, 12, 29).is_last_day_of_quarter(bizdays=True))
        self.assertTrue(Date(2023, 12, 29).is_last_day_of_quarter(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(Date(2023, 12, 31).is_last_day_of_quarter(bizdays=True))
        self.assertFalse(Date(2023, 12, 31).is_last_day_of_quarter(bizdays=True, calendar='ANBIMA'))

    def test_date_is_last_day_of_year(self):
        self.assertFalse(Date(2023, 1, 1).is_last_day_of_year())
        self.assertFalse(Date(2023, 1, 31).is_last_day_of_year())
        self.assertTrue(Date(2023, 12, 31).is_last_day_of_year())
        self.assertTrue(Date(2023, 12, 29).is_last_day_of_year(bizdays=True))
        self.assertTrue(Date(2023, 12, 29).is_last_day_of_year(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(Date(2023, 12, 30).is_last_day_of_year(bizdays=True))
        self.assertFalse(Date(2023, 12, 31).is_last_day_of_year(bizdays=True, calendar='ANBIMA'))

    def test_date_get_first_day_of_week(self):
        self.assertEqual(Date(2023, 1, 1).get_first_day_of_week().__class__, Date)
        self.assertEqual(Date(2023, 1, 1).get_first_day_of_week(), Date(2023, 1, 1))
        self.assertEqual(Date(2023, 1, 5).get_first_day_of_week(), Date(2023, 1, 1))
        self.assertEqual(Date(2023, 1, 7).get_first_day_of_week(bizdays=True), Date(2023, 1, 2))
        self.assertEqual(Date(2023, 1, 1).get_first_day_of_week(bizdays=True, calendar='ANBIMA'), Date(2023, 1, 2))
        self.assertEqual(Date(2023, 4, 3).get_first_day_of_week(), Date(2023, 4, 2))
        self.assertEqual(Date(2023, 4, 3).get_first_day_of_week(bizdays=True), Date(2023, 4, 3))
        self.assertEqual(Date(2023, 4, 3).get_first_day_of_week(bizdays=True, calendar='ANBIMA'), Date(2023, 4, 3))
        self.assertEqual(Date(2023, 5, 4).get_first_day_of_week(bizdays=True, calendar='ANBIMA'), Date(2023, 5, 2))

    def test_date_get_first_day_of_month(self):
        self.assertEqual(Date(2023, 5, 15).get_first_day_of_month().__class__, Date)
        self.assertEqual(Date(2023, 5, 15).get_first_day_of_month(), Date(2023, 5, 1))
        self.assertEqual(Date(2023, 1, 15).get_first_day_of_month(bizdays=True), Date(2023, 1, 2))
        self.assertEqual(Date(2023, 5, 15).get_first_day_of_month(bizdays=True, calendar='ANBIMA'), Date(2023, 5, 2))

    def test_date_is_first_day_of_week(self):
        self.assertTrue(Date(2023, 1, 1).is_first_day_of_week())
        self.assertTrue(Date(2023, 4, 2).is_first_day_of_week())
        self.assertFalse(Date(2023, 1, 1).is_first_day_of_week(bizdays=True))
        self.assertTrue(Date(2018, 1, 2).is_first_day_of_week(bizdays=True, calendar='ANBIMA'))

    def test_date_get_first_day_of_quarter(self):
        self.assertEqual(Date(2023, 1, 15).get_first_day_of_quarter().__class__, Date)
        self.assertEqual(Date(2023, 1, 15).get_first_day_of_quarter(), Date(2023, 1, 1))
        self.assertEqual(Date(2023, 4, 15).get_first_day_of_quarter(), Date(2023, 4, 1))
        self.assertEqual(Date(2023, 4, 15).get_first_day_of_quarter(bizdays=True), Date(2023, 4, 3))
        self.assertEqual(Date(2023, 11, 15).get_first_day_of_quarter(bizdays=True, calendar='ANBIMA'), Date(2023, 10, 2))

    def test_date_get_first_day_of_year(self):
        self.assertEqual(Date(2023, 1, 15).get_first_day_of_year().__class__, Date)
        self.assertEqual(Date(2023, 1, 15).get_first_day_of_year(), Date(2023, 1, 1))
        self.assertEqual(Date(2023, 7, 15).get_first_day_of_year(bizdays=True), Date(2023, 1, 2))
        self.assertEqual(Date(2018, 11, 15).get_first_day_of_year(bizdays=True, calendar='ANBIMA'), Date(2018, 1, 2))

    def test_date_is_first_day_of_month(self):
        self.assertTrue(Date(2023, 1, 1).is_first_day_of_month())
        self.assertTrue(Date(2023, 12, 1).is_first_day_of_month())
        self.assertFalse(Date(2023, 12, 31).is_first_day_of_month())
        self.assertTrue(Date(2023, 1, 2).is_first_day_of_month(bizdays=True))
        self.assertFalse(Date(2023, 1, 1).is_first_day_of_month(bizdays=True))
        self.assertTrue(Date(2018, 1, 2).is_first_day_of_month(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(Date(2018, 1, 1).is_first_day_of_month(bizdays=True, calendar='ANBIMA'))

    def test_date_is_first_day_of_quarter(self):
        self.assertTrue(Date(2023, 1, 1).is_first_day_of_quarter())
        self.assertFalse(Date(2023, 3, 31).is_first_day_of_quarter())
        self.assertTrue(Date(2023, 4, 1).is_first_day_of_quarter())
        self.assertFalse(Date(2023, 6, 30).is_first_day_of_quarter())
        self.assertTrue(Date(2023, 7, 1).is_first_day_of_quarter())
        self.assertFalse(Date(2023, 9, 30).is_first_day_of_quarter())
        self.assertTrue(Date(2023, 10, 1).is_first_day_of_quarter())
        self.assertFalse(Date(2023, 12, 31).is_first_day_of_quarter())
        self.assertTrue(Date(2023, 1, 2).is_first_day_of_quarter(bizdays=True))
        self.assertTrue(Date(2018, 7, 2).is_first_day_of_quarter(bizdays=True))
        self.assertTrue(Date(2018, 1, 2).is_first_day_of_quarter(bizdays=True, calendar='ANBIMA'))

    def test_date_is_first_day_of_year(self):
        self.assertTrue(Date(2023, 1, 1).is_first_day_of_year())
        self.assertTrue(Date(2023, 1, 2).is_first_day_of_year(bizdays=True))
        self.assertTrue(Date(2018, 1, 2).is_first_day_of_year(bizdays=True, calendar='ANBIMA'))
        self.assertFalse(Date(2023, 12, 30).is_first_day_of_year(bizdays=True))
        self.assertFalse(Date(2023, 12, 31).is_first_day_of_year(bizdays=True, calendar='ANBIMA'))

    def test_date_delta(self):
        self.assertEqual(Date(2023, 7, 31).delta(period=5, periodicity='D').__class__, Date)
        self.assertEqual(Date(2023, 7, 31).delta(period=5, periodicity='D'), Date(2023, 8, 5))
        self.assertEqual(Date(2023, 7, 31).delta(period=2, periodicity='W'), Date(2023, 8, 14))
        self.assertEqual(Date(2023, 7, 31).delta(period=1, periodicity='M'), Date(2023, 8, 31))
        self.assertEqual(Date(2023, 7, 31).delta(period=2, periodicity='Y'), Date(2025, 7, 31))
        self.assertEqual(Date(2023, 4, 30).delta(period=1, periodicity='B', calendar='ANBIMA'), Date(2023, 5, 2))

        with self.assertRaisesRegex(ValueError, "Invalid periodicity. Please choose one of the following: D, B, W, M, Y."):
            self.assertEqual(Date(2023, 7, 31).delta(period=5, periodicity='INVALID'), Date(2023, 8, 5))

    def test_date_days_delta(self):
        self.assertEqual(Date(2023, 7, 31).days_delta(5).__class__, Date)
        self.assertEqual(Date(2023, 7, 31).days_delta(5), Date(2023, 8, 5))
        self.assertEqual(Date(2023, 7, 31).days_delta(-5), Date(2023, 7, 26))

    def test_date_bizdays_delta(self):
        self.assertEqual(Date(2023, 7, 31).bizdays_delta(5).__class__, Date)
        self.assertEqual(Date(2023, 7, 31).bizdays_delta(5), Date(2023, 8, 7))
        self.assertEqual(Date(2023, 7, 31).bizdays_delta(-5), Date(2023, 7, 24))
        self.assertEqual(Date(2023, 4, 30).bizdays_delta(1, calendar='ANBIMA'), Date(2023, 5, 2))

        with self.assertRaisesRegex(ValueError, "Value out of range"):
            self.assertEqual(Date(2023, 7, 31).bizdays_delta(50001), Date(2023, 8, 5))

    def test_date_weeks_delta(self):
        self.assertEqual(Date(2023, 7, 31).weeks_delta(5).__class__, Date)
        self.assertEqual(Date(2023, 7, 31).weeks_delta(5), Date(2023, 9, 4))
        self.assertEqual(Date(2023, 7, 31).weeks_delta(-5), Date(2023, 6, 26))

    def test_date_months_delta(self):
        self.assertEqual(Date(2023, 7, 31).months_delta(5).__class__, Date)
        self.assertEqual(Date(2023, 7, 31).months_delta(5), Date(2023, 12, 31))
        self.assertEqual(Date(2023, 7, 31).months_delta(-5), Date(2023, 2, 28))

    def test_date_months_delta_with_leap_year(self):
        self.assertEqual(Date(2020, 3, 31).months_delta(1), Date(2020, 4, 30))
        self.assertEqual(Date(2020, 2, 29).months_delta(1), Date(2020, 3, 29))
        self.assertEqual(Date(2020, 3, 31).months_delta(-1), Date(2020, 2, 29))
        self.assertEqual(Date(2020, 2, 29).months_delta(12), Date(2021, 2, 28))
        self.assertEqual(Date(2020, 2, 29).months_delta(-12), Date(2019, 2, 28))

    def test_date_years_delta(self):
        self.assertEqual(Date(2023, 7, 31).years_delta(5).__class__, Date)
        self.assertEqual(Date(2023, 7, 31).years_delta(5), Date(2028, 7, 31))
        self.assertEqual(Date(2023, 7, 31).years_delta(-5), Date(2018, 7, 31))

    def test_date_years_delta_with_leap_year(self):
        self.assertEqual(Date(2020, 2, 29).years_delta(1), Date(2021, 2, 28))
        self.assertEqual(Date(2020, 2, 29).years_delta(-1), Date(2019, 2, 28))
        self.assertEqual(Date(2020, 2, 29).years_delta(4), Date(2024, 2, 29))
        self.assertEqual(Date(2020, 2, 29).years_delta(-4), Date(2016, 2, 29))

    def test_date_is_business_day(self):
        self.assertEqual(Date(2025, 1, 1).is_business_day(), True)
        self.assertEqual(Date(2025, 1, 1).is_business_day(calendar='ANBIMA'), False)
        self.assertEqual(Date(2025, 1, 20).is_business_day(calendar='NYSE'), False)

    def test_date_nearest_business_day(self):
        self.assertEqual(Date(2025, 1, 1).nearest_business_day(), Date(2025, 1, 1))
        self.assertEqual(Date(2025, 1, 1).nearest_business_day(direction='preceding'), Date(2025, 1, 1))
        self.assertEqual(Date(2025, 1, 1).nearest_business_day(calendar='ANBIMA'), Date(2025, 1, 2))
        self.assertEqual(Date(2025, 1, 20).nearest_business_day(calendar='NYSE'), Date(2025, 1, 21))
        self.assertEqual(Date(2025, 1, 1).nearest_business_day(direction='preceding', calendar='BVMF'), Date(2024, 12, 31))
        self.assertEqual(Date(2025, 4, 21).nearest_business_day(direction='preceding', calendar='BR'), Date(2025, 4, 17))
        self.assertEqual(Date(2025, 4, 18).nearest_business_day(direction='following', calendar='BR'), Date(2025, 4, 22))

        with self.assertRaisesRegex(ValueError, "Invalid direction. Choose 'following' or 'preceding'."):
            self.assertEqual(Date(2025, 1, 1).nearest_business_day(direction='INVALID'), "Invalid direction. Choose 'following' or 'preceding'.")

    def test_date_diff(self):
        self.assertEqual(Date.diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 7, 31), periodicity='D'), 0)
        self.assertEqual(Date.diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 8, 5), periodicity='D'), 5)
        self.assertEqual(Date.diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 8, 14), periodicity='W'), 2)
        self.assertEqual(Date.diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 8, 31), periodicity='M'), 1)
        self.assertEqual(Date.diff(start_date=Date(2023, 7, 31), end_date=Date(2025, 7, 31), periodicity='Y'), 2)
        self.assertEqual(Date.diff(start_date=Date(2023, 4, 30), end_date=Date(2023, 5, 2), periodicity='B', calendar='ANBIMA'), 1)
        self.assertEqual(Date.diff(start_date=date(2023, 4, 30), end_date=date(2023, 5, 2), periodicity='B', calendar='ANBIMA'), 1)

        with self.assertRaisesRegex(ValueError, "Invalid periodicity. Please choose one of the following: D, B, W, M, Y."):
            self.assertEqual(Date.diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 8, 5), periodicity='INVALID'), 5)

    def test_date_days_diff(self):
        self.assertEqual(Date.days_diff(start_date=Date(2023, 8, 5), end_date=Date(2023, 7, 31)), -5)
        self.assertEqual(Date.days_diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 8, 5)), 5)
        self.assertEqual(Date.days_diff(start_date=date(2023, 7, 31), end_date=date(2023, 8, 5)), 5)
        self.assertEqual(Date.days_diff(start_date=Date(2024, 2, 19), end_date=Date(2024, 2, 20)), 1)
        self.assertEqual(Date.days_diff(start_date=Date(2023, 5, 2), end_date=Date(2023, 4, 30)), -2)

    def test_date_bizdays_diff(self):
        self.assertEqual(Date.bizdays_diff(start_date=Date(2023, 8, 7), end_date=Date(2023, 7, 31)), -5)
        self.assertEqual(Date.bizdays_diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 8, 7)), 5)
        self.assertEqual(Date.bizdays_diff(start_date=Date(2023, 5, 2), end_date=Date(2023, 4, 30), calendar='ANBIMA'), -1)
        self.assertEqual(Date.bizdays_diff(start_date=date(2023, 5, 2), end_date=date(2023, 4, 30), calendar='ANBIMA'), -1)
        self.assertEqual(Date.bizdays_diff(start_date=Date(2024, 2, 19), end_date=Date(2024, 2, 20), calendar='ANBIMA'), 1)
        self.assertEqual(Date.bizdays_diff(start_date=Date(2023, 5, 2), end_date=Date(2023, 4, 30), calendar='ANBIMA'), -1)

    def test_date_weeks_diff(self):
        self.assertEqual(Date.weeks_diff(start_date=Date(2023, 9, 4), end_date=Date(2023, 7, 31)), -5)
        self.assertEqual(Date.weeks_diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 9, 4)), 5)
        self.assertEqual(Date.weeks_diff(start_date=date(2023, 7, 31), end_date=date(2023, 9, 4)), 5)

    def test_date_months_diff(self):
        self.assertEqual(Date.months_diff(start_date=Date(2023, 12, 31), end_date=Date(2023, 7, 31)), -5)
        self.assertEqual(Date.months_diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 12, 31)), 5)
        self.assertEqual(Date.months_diff(start_date=Date(2023, 7, 31), end_date=Date(2023, 12, 29)), 4)
        self.assertEqual(Date.months_diff(start_date=date(2023, 7, 31), end_date=date(2023, 12, 29)), 4)

    def test_date_years_diff(self):
        self.assertEqual(Date.years_diff(start_date=Date(2028, 7, 31), end_date=Date(2023, 7, 31)), -5)
        self.assertEqual(Date.years_diff(start_date=Date(2023, 7, 31), end_date=Date(2028, 7, 31)), 5)
        self.assertEqual(Date.years_diff(start_date=Date(2023, 7, 31), end_date=Date(2028, 7, 29)), 4)
        self.assertEqual(Date.years_diff(start_date=date(2023, 7, 31), end_date=date(2028, 7, 29)), 4)

    def test_date_range(self):
        self.assertEqual(Date.range(Date(2023, 1, 1), Date(2023, 1, 1)), [])
        self.assertListEqual(
            Date.range(Date(2021, 12, 23), Date(2022, 1, 2)),
            [
                Date(2021, 12, 23),
                Date(2021, 12, 24),
                Date(2021, 12, 25),
                Date(2021, 12, 26),
                Date(2021, 12, 27),
                Date(2021, 12, 28),
                Date(2021, 12, 29),
                Date(2021, 12, 30),
                Date(2021, 12, 31),
                Date(2022, 1, 1)
            ]
        )

        self.assertEqual(
            Date.range(date(2021, 12, 23), date(2022, 1, 2)),
            [
                Date(2021, 12, 23),
                Date(2021, 12, 24),
                Date(2021, 12, 25),
                Date(2021, 12, 26),
                Date(2021, 12, 27),
                Date(2021, 12, 28),
                Date(2021, 12, 29),
                Date(2021, 12, 30),
                Date(2021, 12, 31),
                Date(2022, 1, 1)
            ]
        )

        self.assertEqual(
            Date.range(Date(2023, 4, 18), Date(2023, 5, 3), periodicity='B', calendar='ANBIMA'),
            [
                Date(2023, 4, 18),
                Date(2023, 4, 19),
                Date(2023, 4, 20),
                Date(2023, 4, 24),
                Date(2023, 4, 25),
                Date(2023, 4, 26),
                Date(2023, 4, 27),
                Date(2023, 4, 28),
                Date(2023, 5, 2)]
        )

    def test_date_range_invalid_periodicity(self):
        with self.assertRaisesRegex(ValueError, 'Invalid periodicity. Please choose one of the following: D, B.'):
            Date.range(Date(2021, 12, 23), Date(2022, 1, 2), periodicity='invalid')

    def test_date_range_out_of_range_date(self):
        with self.assertRaisesRegex(ValueError, 'Value out of range'):
            Date.range(Date(2000, 1, 1), Date(2300, 1, 1))

    def test_date_range_return_correct_type(self):
        start_date, end_date = Date.date_range('custom_range', '', '', None, Date(2023, 9, 28))
        self.assertEqual(start_date.__class__, Date)
        self.assertEqual(end_date.__class__, Date)
        self.assertEqual(start_date, Date(2008, 1, 1))
        self.assertEqual(end_date, Date(2023, 9, 28))

    def test_date_days_range(self):
        self.assertEqual(Date.days_range(Date(2023, 1, 1), Date(2023, 1, 1)), [])
        self.assertListEqual(
            Date.days_range(Date(2021, 12, 23), Date(2022, 1, 2)),
            [
                Date(2021, 12, 23),
                Date(2021, 12, 24),
                Date(2021, 12, 25),
                Date(2021, 12, 26),
                Date(2021, 12, 27),
                Date(2021, 12, 28),
                Date(2021, 12, 29),
                Date(2021, 12, 30),
                Date(2021, 12, 31),
                Date(2022, 1, 1)
            ]
        )

        self.assertListEqual(
            Date.days_range(date(2021, 12, 23), date(2022, 1, 2)),
            [
                Date(2021, 12, 23),
                Date(2021, 12, 24),
                Date(2021, 12, 25),
                Date(2021, 12, 26),
                Date(2021, 12, 27),
                Date(2021, 12, 28),
                Date(2021, 12, 29),
                Date(2021, 12, 30),
                Date(2021, 12, 31),
                Date(2022, 1, 1)
            ]
        )

    def test_date_bizdays_range_with_calendar(self):
        self.assertListEqual(
            Date.bizdays_range(Date(2023, 4, 18), Date(2023, 5, 3), calendar='ANBIMA'),
            [
                Date(2023, 4, 18),
                Date(2023, 4, 19),
                Date(2023, 4, 20),
                Date(2023, 4, 24),
                Date(2023, 4, 25),
                Date(2023, 4, 26),
                Date(2023, 4, 27),
                Date(2023, 4, 28),
                Date(2023, 5, 2)]
        )

        self.assertListEqual(
            Date.bizdays_range(date(2023, 4, 18), date(2023, 5, 3), calendar='ANBIMA'),
            [
                Date(2023, 4, 18),
                Date(2023, 4, 19),
                Date(2023, 4, 20),
                Date(2023, 4, 24),
                Date(2023, 4, 25),
                Date(2023, 4, 26),
                Date(2023, 4, 27),
                Date(2023, 4, 28),
                Date(2023, 5, 2)]
        )

    def test_date_bizdays_range_without_calendar(self):
        self.assertEqual(
            Date.bizdays_range(Date(2023, 4, 18), Date(2023, 5, 3)),
            [
                Date(2023, 4, 18),
                Date(2023, 4, 19),
                Date(2023, 4, 20),
                Date(2023, 4, 21),
                Date(2023, 4, 24),
                Date(2023, 4, 25),
                Date(2023, 4, 26),
                Date(2023, 4, 27),
                Date(2023, 4, 28),
                Date(2023, 5, 1),
                Date(2023, 5, 2)]
        )

    def get_date_parameters(self, range_type='single_date', period='days', nPeriod='10', start_date=Date(2023, 5, 1), end_date=Date(2023, 5, 25)):
        return range_type, period, nPeriod, start_date, end_date

    def test_date_range_single_date(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters()
        cmp = (Date(2023, 5, 25), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_single_date_without_end_date(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(end_date='')
        today = date.today()
        cmp = (today, today)
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_custom_range(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='custom_range')
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        cmp = (Date(2023, 5, 1), Date(2023, 5, 25))
        self.assertEqual(res, cmp)

    def test_date_range_custom_range_without_start_date(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='custom_range', start_date='')
        cmp = (Date(2008, 1, 1), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_n_period_days(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='n_period')
        cmp = (Date(2023, 5, 15), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_n_period_weeks(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='n_period', period='weeks')
        cmp = (Date(2023, 3, 16), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_n_period_months(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='n_period', period='months')
        cmp = (Date(2022, 7, 25), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_n_period_years(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='n_period', period='years')
        cmp = (Date(2013, 5, 25), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_invalid_n_period(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='n_period', period='invalid')
        error_message = 'invalid n period'
        with self.assertRaises(Exception) as error:
            Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertTrue(error_message in str(error.exception))

    def test_date_range_period_to_date_mtd(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='period_to_date', period='mtd')
        cmp = (Date(2023, 5, 1), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_period_to_date_qtd(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='period_to_date', period='qtd')
        cmp = (Date(2023, 4, 1), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_period_to_date_ytd(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='period_to_date', period='ytd')
        cmp = (Date(2023, 1, 1), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_end_period_to_date_ytd(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='end_period_to_date', period='ytd')
        cmp = (Date(2022, 12, 30), Date(2023, 5, 25))
        res = Date.date_range(range_type, period, nPeriod, start_date, end_date)
        self.assertEqual(res, cmp)

    def test_date_range_period_to_date_invalid_period(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='end_period_to_date', period='invalid')
        with self.assertRaisesRegex(ValueError, "invalid period to date"):
            Date.date_range(range_type, period, nPeriod, start_date, end_date)

    def test_date_range_invalid_range_type(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='invalid')
        with self.assertRaisesRegex(ValueError, "invalid range type"):
            Date.date_range(range_type, period, nPeriod, start_date, end_date)

    def test_date_range_invalid_end_date(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='period_to_date', period='mtd', end_date='20230101')
        with self.assertRaisesRegex(ValueError, 'Invalid end_date type'):
            Date.date_range(range_type, period, nPeriod, start_date, end_date)

    def test_date_range_invalid_start_date(self):
        range_type, period, nPeriod, start_date, end_date = self.get_date_parameters(range_type='period_to_date', period='mtd', start_date='20230101')
        with self.assertRaisesRegex(ValueError, 'Invalid start_date type'):
            Date.date_range(range_type, period, nPeriod, start_date, end_date)

    def test_get_date_from_expression_with_normal_day(self):
        self.assertEqual(Date.get_date_from_expression('15th day', 2024, 8), Date(2024, 8, 15))
        self.assertEqual(Date.get_date_from_expression('11th day', 2024, 8), Date(2024, 8, 11))
        self.assertEqual(Date.get_date_from_expression('first day', 2024, 8), Date(2024, 8, 1))
        self.assertEqual(Date.get_date_from_expression('second day', 2024, 8), Date(2024, 8, 2))
        self.assertEqual(Date.get_date_from_expression('last day', 2024, 8), Date(2024, 8, 31))
        self.assertEqual(Date.get_date_from_expression('third day', 2024, 8), Date(2024, 8, 3))
        self.assertEqual(Date.get_date_from_expression('4th day', 2024, 8), Date(2024, 8, 4))
        self.assertEqual(Date.get_date_from_expression('4th day', 2024, 11), Date(2024, 11, 4))
        self.assertEqual(Date.get_date_from_expression('last day', 2024, 11), Date(2024, 11, 30))
        self.assertEqual(Date.get_date_from_expression('15th day', 2024, 11), Date(2024, 11, 15))
        self.assertEqual(Date.get_date_from_expression('28th day', 2024, 2), Date(2024, 2, 28))
        self.assertEqual(Date.get_date_from_expression('last day', 2024, 12), Date(2024, 12, 31))
        self.assertEqual(Date.get_date_from_expression('last day', 2024, 12), Date(2024, 12, 31))
        self.assertEqual(Date.get_date_from_expression('last day', 2020, 2), Date(2020, 2, 29))
        self.assertEqual(Date.get_date_from_expression('last day', 2016, 2), Date(2016, 2, 29))
        self.assertEqual(Date.get_date_from_expression('first day', 2016, 2), Date(2016, 2, 1))
        self.assertEqual(Date.get_date_from_expression('12Th DaY', 2024, 6), Date(2024, 6, 12))
        self.assertEqual(Date.get_date_from_expression('FIRST DaY', 2024, 2), Date(2024, 2, 1))
        self.assertEqual(Date.get_date_from_expression('17TH dAy', 2023, 6), Date(2023, 6, 17))

    def test_get_date_from_expression_with_business_day(self):
        self.assertEqual(Date.get_date_from_expression('13th bizday', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 21))
        self.assertEqual(Date.get_date_from_expression('11th bizday', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 18))
        self.assertEqual(Date.get_date_from_expression('12th bizday', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 19))
        self.assertEqual(Date.get_date_from_expression('last bizday', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 29))
        self.assertEqual(Date.get_date_from_expression('first bizday', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 1))
        self.assertEqual(Date.get_date_from_expression('second bizday', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 4))
        self.assertEqual(Date.get_date_from_expression('third bizday', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 5))
        self.assertEqual(Date.get_date_from_expression('first bizday', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 2))
        self.assertEqual(Date.get_date_from_expression('second bizday', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 3))
        self.assertEqual(Date.get_date_from_expression('third bizday', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 6))
        self.assertEqual(Date.get_date_from_expression('4th bizday', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 7))
        self.assertEqual(Date.get_date_from_expression('second bizday', 2024, 2, calendar='ANBIMA'), Date(2024, 2, 2))
        self.assertEqual(Date.get_date_from_expression('third bizday', 2024, 2, calendar='ANBIMA'), Date(2024, 2, 5))
        self.assertEqual(Date.get_date_from_expression('7th bizday', 2024, 2, calendar='ANBIMA'), Date(2024, 2, 9))
        self.assertEqual(Date.get_date_from_expression('8th bizday', 2024, 2, calendar='ANBIMA'), Date(2024, 2, 14))
        self.assertEqual(Date.get_date_from_expression('8th bizday', 2024, 2, calendar='BR'), Date(2024, 2, 14))
        self.assertEqual(Date.get_date_from_expression('21th bizday', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 31))
        self.assertEqual(Date.get_date_from_expression('12TH BIZDAY', 2024, 6, calendar='ANBIMA'), Date(2024, 6, 18))
        self.assertEqual(Date.get_date_from_expression('THIrd BIZDAY', 2024, 6, calendar='ANBIMA'), Date(2024, 6, 5))
        self.assertEqual(Date.get_date_from_expression('7TH BiZdAY', 2024, 6, calendar='ANBIMA'), Date(2024, 6, 11))
        self.assertEqual(Date.get_date_from_expression('FiRsT bizDAY', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 1))
        self.assertEqual(Date.get_date_from_expression('SECOnD BizDAY', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 4))

    def test_get_date_from_expression_with_invalid_date_for_business_day(self):
        with self.assertRaisesRegex(ValueError, 'Invalid day: nonday'):
            Date.get_date_from_expression('10th nonday', 2024, 11, calendar='ANBIMA')

        with self.assertRaisesRegex(DateError, 'Invalid Business Day'):
            Date.get_date_from_expression('25th bizday', 2024, 11, calendar='ANBIMA')

        with self.assertRaisesRegex(DateError, 'Invalid Business Day'):
            Date.get_date_from_expression('31th bizday', 2024, 1, calendar='ANBIMA')

    def test_get_date_from_expression_with_position_bizdays(self):
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2024, 8, 19, calendar='ANBIMA'), Date(2024, 8, 16))
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2024, 11, 21, calendar='ANBIMA'), Date(2024, 11, 19))
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2023, 11, 21, calendar='ANBIMA'), Date(2023, 11, 20))
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2022, 11, 21, calendar='ANBIMA'), Date(2022, 11, 18))
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2022, 11, 21), Date(2022, 11, 18))
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2024, 5, 2, calendar='ANBIMA'), Date(2024, 4, 30))
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2024, 11, 18, calendar='ANBIMA'), Date(2024, 11, 14))
        self.assertEqual(Date.get_date_from_expression('previous bizday', 2024, 5, 31, calendar='ANBIMA'), Date(2024, 5, 29))
        self.assertEqual(Date.get_date_from_expression('next bizday', 2024, 4, 30, calendar='ANBIMA'), Date(2024, 5, 2))
        self.assertEqual(Date.get_date_from_expression('next bizday', 2024, 11, 19, calendar='ANBIMA'), Date(2024, 11, 21))
        self.assertEqual(Date.get_date_from_expression('next bizday', 2024, 5, 29, calendar='ANBIMA'), Date(2024, 5, 31))
        self.assertEqual(Date.get_date_from_expression('next bizday', 2024, 11, 14, calendar='ANBIMA'), Date(2024, 11, 18))
        self.assertEqual(Date.get_date_from_expression('next bizday', 2024, 8, 23, calendar='ANBIMA'), Date(2024, 8, 26))
        self.assertEqual(Date.get_date_from_expression('next bizday', 2024, 12, 24, calendar='ANBIMA'), Date(2024, 12, 26))
        self.assertEqual(Date.get_date_from_expression('PREvIoUs biZday', 2024, 8, 19, calendar='ANBIMA'), Date(2024, 8, 16), Date.get_date_from_expression('previous bizday', 2024, 8, 19, calendar='ANBIMA'))
        self.assertEqual(Date.get_date_from_expression('NEXT bizday', 2024, 12, 24, calendar='ANBIMA'), Date.get_date_from_expression('next bizday', 2024, 12, 24, calendar='ANBIMA'))
        self.assertEqual(Date.get_date_from_expression('NExT biZdAy', 2024, 5, 29, calendar='ANBIMA'), Date(2024, 5, 31))
        self.assertEqual(Date.get_date_from_expression('PREVIOUS bizday', 2024, 11, 18, calendar='ANBIMA'), Date(2024, 11, 14))
        self.assertEqual(Date.get_date_from_expression('PREVIOUS BIZDAY', 2022, 11, 21, calendar='ANBIMA'), Date.get_date_from_expression('pReViOus bIzDaY', 2022, 11, 21, calendar='ANBIMA'))

    def test_get_date_from_expression_with_bizdays_after_nth_day(self):
        self.assertEqual(Date.get_date_from_expression('first bizday after 15th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 18))
        self.assertEqual(Date.get_date_from_expression('second bizday after 15th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 19))
        self.assertEqual(Date.get_date_from_expression('first bizday after 19th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 21))
        self.assertEqual(Date.get_date_from_expression('second bizday after 18th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 21))
        self.assertEqual(Date.get_date_from_expression('first bizday after 10th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 11))
        self.assertEqual(Date.get_date_from_expression('first bizday after 8th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 11))
        self.assertEqual(Date.get_date_from_expression('third bizday after 7th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 12))
        self.assertEqual(Date.get_date_from_expression('first bizday after 14th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 18))
        self.assertEqual(Date.get_date_from_expression('first bizday after 17th day', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 20))
        self.assertEqual(Date.get_date_from_expression('second bizday after 17th day', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 21))
        self.assertEqual(Date.get_date_from_expression('SECOnd biZDAy afTer 17TH DAy', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 21))
        self.assertEqual(Date.get_date_from_expression('firST BIZDAY AFTER 17Th dAy', 2024, 5, calendar='ANBIMA'), Date(2024, 5, 20))

    def test_get_date_from_expression_with_bizdays_before_nth_day(self):
        self.assertEqual(Date.get_date_from_expression('first bizday before 14th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 13))
        self.assertEqual(Date.get_date_from_expression('first bizday before 21th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 19))
        self.assertEqual(Date.get_date_from_expression('first bizday before 22th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 21))
        self.assertEqual(Date.get_date_from_expression('second bizday before 22th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 19))
        self.assertEqual(Date.get_date_from_expression('third bizday before 23th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 19))
        self.assertEqual(Date.get_date_from_expression('first bizday before 18th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 14))
        self.assertEqual(Date.get_date_from_expression('first bizday before 2nd day', 2024, 5, calendar='ANBIMA'), Date(2024, 4, 30))
        self.assertEqual(Date.get_date_from_expression('FIrsT BIZDAY BEFore 2Nd DAY', 2024, 5, calendar='ANBIMA'), Date(2024, 4, 30))
        self.assertEqual(Date.get_date_from_expression('firST bizdaY beforE 2nd day', 2024, 5, calendar='ANBIMA'), Date(2024, 4, 30))

    def test_get_date_from_expression_with_bizdays_containing_invalid_expression(self):
        with self.assertRaisesRegex(ValueError, 'Invalid expression: next, the day reference must be either "before" or "after"'):
            Date.get_date_from_expression('first bizday next 15th day', 2024, 11, calendar='ANBIMA')

        with self.assertRaisesRegex(ValueError, 'Invalid expression, please verify the day type or the week days'):
            Date.get_date_from_expression('first sae bizday before 15th day', 2024, 11, calendar='ANBIMA')

        with self.assertRaisesRegex(ValueError, 'Invalid expression: first fri bizday before 15th normal day'):
            Date.get_date_from_expression('first fri bizday before 15th normal day', 2024, 11, calendar='ANBIMA')

        with self.assertRaisesRegex(ValueError, 'Invalid nth: before'):
            Date.get_date_from_expression('before 15th', 2024, 11, calendar='ANBIMA')

        with self.assertRaisesRegex(ValueError, 'Invalid Written Ordinal: nth'):
            Date.get_date_from_expression('nth wed bizday after 15th day', 2024, 11, calendar='ANBIMA')

        with self.assertRaisesRegex(ValueError, 'Invalid Written Ordinal: fourth'):
            Date.get_date_from_expression('fourth bizday before 14th day', 2024, 11, calendar='ANBIMA')

    def test_get_date_from_expression_with_bizdays_after_nth_day_and_day_name(self):
        self.assertEqual(Date.get_date_from_expression('first wed bizday after 15th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 27))
        self.assertEqual(Date.get_date_from_expression('first wed bizday after 28th day', 2024, 11, calendar='ANBIMA'), Date(2024, 12, 4))
        self.assertEqual(Date.get_date_from_expression('first fri bizday after 28th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 29))
        self.assertEqual(Date.get_date_from_expression('first mon bizday after 28th day', 2024, 11, calendar='ANBIMA'), Date(2024, 12, 2))
        self.assertEqual(Date.get_date_from_expression('first fri bizday after 14th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 22))
        self.assertEqual(Date.get_date_from_expression('first thu bizday after 29th day', 2024, 5, calendar='ANBIMA'), Date(2024, 6, 6))
        self.assertEqual(Date.get_date_from_expression('first wed bizday after 19th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 27))
        self.assertEqual(Date.get_date_from_expression('first fri bizday after 14th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 22))
        self.assertEqual(Date.get_date_from_expression('first thu bizday after 10th day', 2024, 8, calendar='ANBIMA'), Date(2024, 8, 15))
        self.assertEqual(Date.get_date_from_expression('first tue bizday after 29th day', 2024, 9, calendar='ANBIMA'), Date(2024, 10, 1))
        self.assertEqual(Date.get_date_from_expression('first wed bizday after 5th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 6))
        self.assertEqual(Date.get_date_from_expression('second wed bizday after 5th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 13))
        self.assertEqual(Date.get_date_from_expression('third wed bizday after 5th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 27))
        self.assertEqual(Date.get_date_from_expression('THIRD WED BIZDAY AFTER 5Th DAY', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 27))
        self.assertEqual(Date.get_date_from_expression('first WeD BIZday afTer 5th Day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 6))
        self.assertEqual(Date.get_date_from_expression('first wed bizday after 27th day', 2024, 12, calendar='ANBIMA'), Date(2025, 1, 8))
        self.assertEqual(Date.get_date_from_expression('second wed bizday after 27th day', 2024, 12, calendar='ANBIMA'), Date(2025, 1, 15))

    def test_get_date_from_expression_with_bizdays_before_nth_day_and_day_name(self):
        self.assertEqual(Date.get_date_from_expression('first wed bizday before 15th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 13))
        self.assertEqual(Date.get_date_from_expression('second wed bizday before 15th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 6))
        self.assertEqual(Date.get_date_from_expression('first wed bizday before 22th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 13))
        self.assertEqual(Date.get_date_from_expression('second wed bizday before 22th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 6))
        self.assertEqual(Date.get_date_from_expression('first wed bizday before 8th day', 2024, 5, calendar='ANBIMA'), Date(2024, 4, 24))
        self.assertEqual(Date.get_date_from_expression('first fri bizday before 16th day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 8))
        self.assertEqual(Date.get_date_from_expression('second WED bizday bEfOrE 22TH day', 2024, 11, calendar='ANBIMA'), Date(2024, 11, 6))
        self.assertEqual(Date.get_date_from_expression('FIRSt fRI bIZDAY BEFore 16TH dAY', 2024, 11, calendar='ANBIMA'), Date.get_date_from_expression('first fri bizday before 16th day', 2024, 11, calendar='ANBIMA'))
        self.assertEqual(Date.get_date_from_expression('second wed bizday before 5th day', 2024, 1, calendar='ANBIMA'), Date(2023, 12, 27))
        self.assertEqual(Date.get_date_from_expression('third wed bizday before 5th day', 2024, 1, calendar='ANBIMA'), Date(2023, 12, 20))



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.config import settings
from everysk.core.fields import ListField, StrField, FloatField, ChoiceField
from everysk.core.exceptions import SDKValueError
from everysk.sdk.entities.base import BaseEntity, ScriptMetaClass
from everysk.sdk.entities.query import Query
from everysk.sdk.entities.script import Script
from everysk.sdk.entities.fields import CurrencyField, EntityNameField, EntityDescriptionField, EntityTagsField


###############################################################################
#   CustomIndex Class Implementation
###############################################################################
class CustomIndex(BaseEntity, metaclass=ScriptMetaClass):
    """
    This class represents a custom index entity object.

    Attributes:
        script (Script): The script object associated to the entity.
        _orderable_attributes (ListField): The list of orderable attributes.
        symbol (StrField): The symbol of the custom index.
        id (StrField): The id of the custom index.
        data (ListField): The data of the custom index.

        periodicity (ChoiceField): The periodicity of the custom index.
        data_type (ChoiceField): The data type of the custom index.
        currency (CurrencyField): The currency of the custom index.
        base_price (FloatField): The base price of the custom index.

        description (EntityDescriptionField): The description of the custom index.
        name (EntityNameField): The name of the custom index.
        tags (EntityTagsField): The tags of the custom index.

        created_on (DateTimeField): The creation date of the custom index.
        updated_on (DateTimeField): The last update date of the custom index.
        version (IntField): The version of the custom index.

    Example:
        >>> from everysk.sdk.entities.custom_index.base import CustomIndex
        >>> custom_index = CustomIndex(
            symbol='CUSTOM:INDEX',
            data=[...],
            periodicity='M',
            data_type='PRICE',
            currency='USD',
            base_price=100,
            description='Custom index description',
            name='Custom index name',
            tags=['tag1', 'tag2']
        )
        >>> custom_index.save()
    """

    script: Script
    _orderable_attributes = ListField(default=['created_on', 'updated_on', 'name'], readonly=True)

    symbol = StrField(regex=settings.CUSTOM_INDEX_SYMBOL_REGEX, min_size=settings.CUSTOM_INDEX_SYMBOL_MIN_SIZE, max_size=settings.CUSTOM_INDEX_SYMBOL_MAX_SIZE, required_lazy=True, empty_is_none=True)

    data = ListField(min_size=settings.CUSTOM_INDEX_MIN_DATA_BLOB, required_lazy=True)

    periodicity = ChoiceField(default=None, choices=(settings.CUSTOM_INDEX_PERIODICITY_MONTHLY, settings.CUSTOM_INDEX_PERIODICITY_DAILY, None), required_lazy=True)
    data_type = ChoiceField(default=None, choices=(settings.CUSTOM_INDEX_DATA_TYPE_PRICE, settings.CUSTOM_INDEX_DATA_TYPE_RETURN, settings.CUSTOM_INDEX_DATA_TYPE_RETURN_100, None), required_lazy=True)
    currency = CurrencyField()
    base_price = FloatField(min_size=settings.CUSTOM_INDEX_BASE_PRICE_MIN_VAL, max_size=settings.CUSTOM_INDEX_BASE_PRICE_MAX_VAL)

    description = EntityDescriptionField()
    name = EntityNameField()
    tags = EntityTagsField()

    @property
    def id(self) -> str:
        """
        Returns the id of the custom index. The id is a alias for the symbol.

        Returns:
            str: The id of the custom index.

        Example:
            >>> custom_index = CustomIndex(symbol='CUSTOM:INDEX')
            >>> custom_index.id
            'CUSTOM:INDEX'
        """
        return self.symbol

    @id.setter
    def id(self, value: str) -> None:
        """
        Sets the id of the custom index. The id is a alias for the symbol.

        Args:
            value (str): The id of the custom index.

        Example:
            >>> custom_index = CustomIndex()
            >>> custom_index.id = 'CUSTOM:INDEX'
            >>> custom_index.symbol
            'CUSTOM:INDEX'
            >>> custom_index.id
            'CUSTOM:INDEX'
        """
        self.symbol = value

    def validate(self) -> bool:
        """
        This method validates the entity object and raises an exception if it is not
        valid. The validation is performed by calling the `validate` method of each field
        of the entity.

        Args:
            self (Self): The entity object to be validated.

        Raises:
            FieldValueError: If the entity object is not valid.
            RequiredFieldError: If a required field is missing.

        Returns:
            bool: True if the entity object is valid, False otherwise.

        Example:
            >>> custom_index = CustomIndex(symbol='CUSTOM:INDEX', name='Custom index name', ...)
            >>> custom_index.validate()
            True
        """
        return self.get_response(self_obj=self)

    def _pre_validate(self):
        pass

    def _pos_validate(self):
        self.created_on = None
        self.updated_on = None

    def _check_query(self, query: Query) -> bool:
        """
        Check the query object.

        Args:
            query (Query): The query object.

        Returns:
            Query: The query object.
        """
        return True

    def _check_entity_to_query(self) -> bool:
        """
        Check the entity object to query.

        Returns:
            bool: True if the entity object is valid.
        """
        if self.name and self.tags:
            raise SDKValueError("Can't filter by Name and Tags at the same time")

        return True

    def _mount_query(self, query: Query) -> Query:
        """
        Mount the query object.

        Args:
            query (Query): The query object.

        Returns:
            Query: The query object.
        """
        if self.name is not None:
            query = query.where('name', self.name)
        if self.tags:
            query = query.where('tags', self.tags)

        return query

    @staticmethod
    def get_id_prefix() -> str:
        """
        Returns the prefix of the custom index id field value.

        Returns:
            str: The prefix of the custom index id field value.

        Example:
            >>> from everysk.sdk.custom_index.base import CustomIndex
            >>> CustomIndex.get_id_prefix()
            'CUSTOM:'
        """
        return settings.CUSTOM_INDEX_SYMBOL_PREFIX

    @classmethod
    def modify_many(cls, entity_id_list: list[str], overwrites: dict | list[dict]) -> list:
        """
        The custom index does not support the modify_many method.
        """
        raise NotImplementedError()

    @classmethod
    def clone_many(cls, entity_id_list: list[str], overwrites: dict | list[dict]) -> list:
        """
        The custom index does not support the clone_many method.
        """
        raise NotImplementedError()



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from typing import Any

from everysk.config import settings
from everysk.core.fields import DateField, ListField, StrField, BoolField, ChoiceField
from everysk.core.datetime import Date
from everysk.sdk.base import BaseSDK


###############################################################################
#   MarketData Class Implementation
###############################################################################
class MarketData(BaseSDK):

    date = DateField(default=Undefined)
    start_date = DateField(default=None)
    end_date = DateField(default=None)
    ticker_list = ListField(default=None)
    ticker_type = ChoiceField(default=None, choices=settings.ENGINES_MARKET_DATA_TICKER_TYPES)
    projection = StrField(default=None)
    nearest = BoolField(default=False)
    real_time = BoolField(default=False)

    def get_security(
        self,
        date: Date = Undefined,
        ticker_list: list[str] = Undefined,
        ticker_type: list[str] = Undefined,
        projection: str = Undefined,
        nearest: bool = Undefined
    ) -> dict:
        """
        Get security data.

        Args:
            date (Date): The date.
            ticker_list (list[str]): The ticker list.
            ticker_type (list[str]): The ticker type.
            projection (str): The projection.
            nearest (bool): The nearest flag.

        Returns:
            dict: The security data.
        """
        return self.get_response(self_obj=self, params={'date': date, 'ticker_list': ticker_list, 'ticker_type': ticker_type, 'projection': projection, 'nearest': nearest})

    def get_historical(
        self,
        date: Date = Undefined,
        start_date: Date = Undefined,
        end_date: Date = Undefined,
        ticker_list: list[str] = Undefined,
        ticker_type: str = Undefined,
        projection: str = Undefined,
        nearest: bool = Undefined,
        real_time: bool = Undefined
    ) -> dict:
        """
        Get historical data.

        Args:
            date (Date): The date.
            start_date (Date): The start date.
            end_date (Date): The end date.
            ticker_list (list[str]): The ticker list.
            ticker_type (str): The ticker type.
            projection (str): The projection.

        Returns:
            dict: The historical data.
        """
        return self.get_response(self_obj=self, params={'date': date, 'start_date': start_date, 'end_date': end_date, 'ticker_list': ticker_list, 'ticker_type': ticker_type, 'projection': projection, 'nearest': nearest, 'real_time': real_time})

    def search(
        self,
        conditions: list[list[str, str, Any]],
        fields: list[str] | None = None,
        order_by: str | None = None,
        limit: int | None = Undefined,
        date: str | Date = Undefined,
        path: str = ''
    ) -> list[dict]:
        """
        Search assets via Market Data Beta with dynamic filters.

        Args:
            *conditions: Each condition is a list or tuple with (field, operator, value). Example: ('cnpj_fundo', '=', '56903183000100')
            fields (list[str]): List of fields to include in the response.
            order_by (str): Field to order the results by. Prefix with '-' for descending order (e.g., '-columnA').
            limit (int): Limit the number of results.
            date (str | Date): The date to search for.
            path (str): The path to search for.

        Returns:
            list[dict]: The search results.

        """
        return self.get_response(self_obj=self, params={'conditions': conditions, 'fields': fields, 'order_by': order_by, 'limit': limit, 'date': date, 'path': path})

    def get_currencies(self) -> dict:
        """
        Get the currencies information from the Market Data engine.

        Returns:
            dict: A dictionary with the currencies information.

        Example:
            >>> from everysk.sdk.engines import MarketData

            >>> market_data = MarketData()
            >>> currencies = market_data.currencies()
            >>> currencies
            {
                'base_currencies': [
                    ["AED", "Uae Dirham/Us Dollar Fx Spot Rate"],
                    ["USD", "Us Dollar/Us Dollar Fx Spot Rate"],
                    ...
                ],
                'crypto_currencies': [
                    ["BTC", "Bitcoin"],
                    ["ETH", "Ethereum"],
                    ...
                ],
            }

        """
        return self.get_response(self_obj=self)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
from concurrent import futures
from contextvars import ContextVar
from sys import version_info
from time import sleep
from everysk.core import threads
from everysk.core.datetime import DateTime
from everysk.core.unittests import TestCase, mock, skipUnless


user = ContextVar('username', default='default_user')
def fake_func_context_var():
    return user.get()

def div(v1: int, v2: int) -> int:
    sleep(0.1)
    return v1 / v2

class ThreadTestCase(TestCase):

    def test_results(self):
        results = []
        t1 = threads.Thread(target=div, args=(1, 2))
        t2 = threads.Thread(target=div, args=(1, 4))
        t1.start()
        t2.start()
        results = [t1.join(), t2.join()]
        self.assertEqual(len(results), 2)
        self.assertIn(0.25, results)
        self.assertIn(0.5, results)

    @mock.patch.object(threads, 'log')
    def test_error_3_11(self, log: mock.MagicMock):
        t1 = threads.Thread(target=div, args=(1, 0))
        t1.start()
        t1.join()
        log.error.assert_called_once_with('Thread execution error -> target: %s', 'div', extra={'labels': {'args': (1, 0), 'kwargs': {}}})

    def test_context_var(self):
        user.set('new_user')
        thread = threads.Thread(target=fake_func_context_var)
        thread.start()
        result = thread.join()
        self.assertEqual(result, 'new_user')


def fake_func():
    sleep(1)
    return DateTime.now()

def fake_func_error():
    raise KeyError('MyKey')


class ThreadPoolTestCase(TestCase):

    def test_results(self):
        pool = threads.ThreadPool()
        pool.add(target=div, args=(1, 2))
        pool.add(target=div, args=(1, 4))
        pool.wait()
        self.assertEqual(len(pool.results), 2)
        self.assertIn(0.25, pool.results)
        self.assertIn(0.5, pool.results)

    def test_concurrency(self):
        pool = threads.ThreadPool(2)
        pool.add(target=fake_func)
        pool.add(target=fake_func)
        pool.wait()
        # The second call could finish before, so we need to get the min and max to check the delta
        min_value = min(pool.results)
        max_value = max(pool.results)
        # They run in the same second so the delta will be between 0 and 1
        delta = max_value - min_value
        self.assertLess(delta.total_seconds(), 1)
        self.assertGreater(delta.total_seconds(), 0)

    def test_concurrency_with_one_thread(self):
        pool = threads.ThreadPool(1)
        pool.add(target=fake_func)
        pool.add(target=fake_func)
        pool.wait()
        delta = pool.results[1] - pool.results[0]
        # The second run need to wait 1 second so the delta will be between 1 and 2
        self.assertLess(delta.total_seconds(), 2)
        self.assertGreater(delta.total_seconds(), 1)

    @mock.patch.object(threads.log, 'error')
    def test_silent_thread_3_11(self, error: mock.MagicMock):
        pool = threads.ThreadPool(concurrency=1, silent=True)
        pool.add(target=fake_func_error)
        pool.wait()
        error.assert_called_once_with('Thread execution error -> target: %s', 'fake_func_error', extra={'labels': {'args': (), 'kwargs': {}}})

    @mock.patch.object(threads.log, 'error')
    def test_raise_thread(self, error: mock.MagicMock):
        pool = threads.ThreadPool(concurrency=1, silent=False)
        pool.add(target=fake_func_error)
        with self.assertRaisesRegex(KeyError, 'MyKey'):
            pool.wait()
        error.assert_not_called()

    def test_context_var(self):
        user.set('new_user')
        pool = threads.ThreadPool(concurrency=1)
        pool.add(target=fake_func_context_var)
        pool.wait()
        self.assertListEqual(pool.results, ['new_user'])



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.fields import TupleField, IntField, ListField


ENGINES_EXPRESSION_DEFAULT_DATA_TYPES = TupleField(default=('cpp_var', 'str_var'), readonly=True)
ENGINES_CACHE_EXECUTION_EXPIRATION_TIME = IntField(default=14400, readonly=True)
ENGINES_MARKET_DATA_TICKER_TYPES = ListField(default=('everysk_symbol', 'everysk_id', None), readonly=True)



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from hashlib import sha1

from everysk.core.fields import BoolField, ChoiceField, DictField, StrField
from everysk.core.http import HttpPOSTConnection, httpx
from everysk.core.redis import RedisCacheCompressed


###############################################################################
#   Slack Class Implementation
###############################################################################
class Slack(HttpPOSTConnection):
    ## Private attributes
    _key_prefix = 'everysk-core-slack'
    _cache = RedisCacheCompressed()
    _cache_timeout = 60 * 1 # 1 minute
    _color_map = DictField(default={'danger': '#a90a05', 'success': '#138138', 'warning': '#e9932d'}, readonly=True)
    is_json = BoolField(default=True, readonly=True)

    ## Public attributes
    color = ChoiceField(default=None, choices=(None, 'danger', 'success', 'warning'))
    message = StrField(default=None, required=True)
    title = StrField(default=None, required=True)

    def get_payload(self) -> dict:
        """
        Convert all key/value on self to a dict object and apply some conversions.
        """
        return {
            'attachments': [{
                'color': self._color_map.get(self.color, '#000000'),
                'blocks': [
                    {
                        'type': 'header',
                        'text': {
                            'type': 'plain_text',
                            'text': self.title,
                            'emoji': True
                        }
                    },
                    {
                        'type': 'divider'
                    },
                    {
                        'type': 'section',
                        'text': {
                            'type': 'mrkdwn',
                            'text': self.message
                        }
                    },
                    {
                        'type': 'divider'
                    }
                ]
            }]
        }

    def is_title_cached(self) -> bool:
        """
        Check if the title is already in the cache and set it if not.
        """
        title_key = self.make_title_key()
        cached_title = self._cache.get(title_key)
        if cached_title:
            return True

        # Keep the title in the cache for 1 minute
        self._cache.set(title_key, True, self._cache_timeout)
        return False

    def is_message_cached(self) -> bool:
        """
        Check if the message is already in the cache and set it if not.
        """
        message_key = self.make_message_key()
        cached_message = self._cache.get(message_key)
        if cached_message:
            return True

        # Keep the message in the cache for 1 minute
        self._cache.set(message_key, True, self._cache_timeout)
        return False

    def make_title_key(self) -> str:
        """
        Create a key for the title in the cache.
        """
        if isinstance(self.title, str):
            value = self.title.encode('utf-8')
        else:
            value = self.title

        return f'{self._key_prefix}-{sha1(value).hexdigest()}'

    def make_message_key(self) -> str:
        """
        Create a key for the message in the cache.
        """
        if isinstance(self.message, str):
            value = self.message.encode('utf-8')
        else:
            value = self.message

        return f'{self._key_prefix}-{sha1(value).hexdigest()}'

    def can_send(self) -> bool:
        """
        Check if the title and message are already in the cache.
        If they are, the message will not be sent because it is a duplicate.
        """
        return not self.is_title_cached() and not self.is_message_cached()

    def send(self) -> httpx.Response | None:
        """ Sends the message to the Slack Channel """
        if self.can_send():
            return self.get_response()

        return None



###############################################################################
#
# (C) Copyright 2023 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
# pylint: disable=protected-access
from concurrent.futures import ThreadPoolExecutor
from os import environ
from time import sleep
from uuid import uuid4
from redis import exceptions, client # pylint: disable=import-self
from everysk.core import redis as redis_module
from everysk.core.redis import (
    RedisCache, RedisCacheCompressed, RedisChannel, RedisClient,
    RedisEmptyListError, RedisList, RedisLock, cache
)
from everysk.core.unittests import TestCase, mock


@cache(timeout=1)
def sum(a: int, b: int) -> int: # pylint: disable=redefined-builtin
    return a + b


class CacheDecoratorTestCase(TestCase):

    def setUp(self) -> None:
        self.obj = RedisCache()
        self.obj.flush_all()

    def test_timeout_value_not_int(self):
        with self.assertRaises(ValueError) as context:
            cache(timeout='1')

        self.assertEqual(str(context.exception), 'Timeout must be an integer greater than 0.')

    def test_timeout_value_zero(self):
        with self.assertRaises(ValueError) as context:
            cache(timeout=0)

        self.assertEqual(str(context.exception), 'Timeout must be an integer greater than 0.')

    def test_timeout_value_negative(self):
        with self.assertRaises(ValueError) as context:
            cache(timeout=-1)

        self.assertEqual(str(context.exception), 'Timeout must be an integer greater than 0.')

    def test_cache_decorator(self):
        self.assertEqual(sum(1, 1), 2)
        self.assertEqual(sum(1, 1), 2)
        self.assertListEqual(
            self.obj.connection.keys(),
            [b'sum:e718fe8edf8a7c2917cd7444286b7048042a74abf5208f1d70933c08dbd6b7e9']
        )
        self.assertDictEqual(sum.info, {'hits': 1, 'misses': 1})
        sleep(1.1)
        self.assertEqual(sum(1, 1), 2)
        self.assertDictEqual(sum.info, {'hits': 1, 'misses': 2})

    def test_cache_decorator_clear(self):
        self.obj.set('key', 1)
        self.assertEqual(sum(1, 1), 2)
        self.assertEqual(sum(1, 2), 3)
        self.assertEqual(sum(1, 3), 4)
        self.assertListEqual(
            sorted(self.obj.connection.keys()),
            [
                b'2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683',
                b'sum:3a316d6d3226f84c1e46e4447fa8d5fd800bff4a1bc6498152523cd4a602b69b',
                b'sum:4e3970018fa500922ce5c087f84c734d7b47b0651da2f2e17226801885838032',
                b'sum:e718fe8edf8a7c2917cd7444286b7048042a74abf5208f1d70933c08dbd6b7e9',
            ]
        )
        sum.clear()
        self.assertListEqual(self.obj.connection.keys(), [b'2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'])


class RedisClientTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None

    @mock.patch.object(redis_module, 'log')
    def test_environ(self, log: mock.MagicMock):
        environ['REDIS_HOST'] = '0.1.0.1'
        environ['REDIS_PORT'] = '1234'
        redis = RedisClient()
        self.assertEqual(redis.host, environ['REDIS_HOST'])
        self.assertEqual(redis.port, int(environ['REDIS_PORT']))
        del environ['REDIS_HOST']
        del environ['REDIS_PORT']
        log.debug.assert_called_once_with('Connecting on Redis(%s).....', redis.host)

    @mock.patch.object(redis_module, 'log')
    def test_connection_property(self, log: mock.MagicMock):
        redis = RedisClient()
        self.assertEqual(redis.connection, redis._connection)
        self.assertEqual(redis.connection, RedisClient._connection)
        log.debug.assert_called_once()

    @mock.patch.object(redis_module, 'log')
    def test_hash_key(self, log: mock.MagicMock):
        redis = RedisClient()
        self.assertEqual(
            redis.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        # Check again to guarantee the same hash
        self.assertEqual(
            redis.get_hash_key('a'),
            'ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb'
        )
        log.debug.assert_called_once()

    @mock.patch.object(redis_module, 'log')
    def test_connect(self, log: mock.MagicMock):
        redis = RedisClient()
        log.debug.assert_called_once()
        RedisClient._connection = None
        self.assertIsNone(RedisClient._connection)
        redis._connect()
        self.assertIsNotNone(RedisClient._connection)

    @mock.patch.object(RedisClient, '_connect')
    def test_connection_error_connection(self, _connect: mock.MagicMock):
        with mock.patch.object(RedisClient, '_connection', spec=redis_module.Redis) as _connection:
            _connection.ping.side_effect = exceptions.ConnectionError
            connection = RedisClient().connection # pylint: disable=unused-variable
            _connection.ping.assert_called_once_with()
        _connect.assert_called_once_with()

    @mock.patch.object(redis_module.log, 'debug', mock.MagicMock)
    def test_flush_all(self):
        cli = RedisClient()
        # Generate a random key to test
        key = uuid4().hex
        cli.connection.set(key, 'Flush-test')
        cli.flush_all()
        self.assertIsNone(cli.connection.get(key))


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisCacheTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None

    def test_get_hash_key(self):
        obj = RedisCache()
        hash_key = '2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        self.assertEqual(obj.get_hash_key('key'), hash_key)
        # Must always be the same value
        self.assertEqual(obj.get_hash_key('key'), hash_key)

    def test_get_set_delete(self):
        obj = RedisCache()
        obj.delete('key')
        self.assertIsNone(obj.get('key'))
        obj.set('key', 'My value.')
        self.assertEqual(obj.get('key'), b'My value.')
        obj.set('key', 1)
        self.assertEqual(obj.get('key'), b'1')
        obj.set('key', b'1.1')
        self.assertEqual(obj.get('key'), b'1.1')
        obj.delete('key')
        self.assertIsNone(obj.get('key'))

    def test_set_error(self):
        obj = RedisCache()
        self.assertRaises(ValueError, obj.set, 'key', ())
        self.assertRaises(ValueError, obj.set, 'key', [])
        self.assertRaises(ValueError, obj.set, 'key', {})
        self.assertRaises(ValueError, obj.set, 'key', obj)

    def test_prefix(self):
        key = b'prefix:2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        obj = RedisCache(prefix='prefix')
        self.assertNotIn(key, obj.connection.keys())
        obj.set('key', 1)
        self.assertIn(key, obj.connection.keys())
        obj.delete('key')
        self.assertNotIn(key, obj.connection.keys())

    def test_delete_prefix(self):
        keys = [
            b'other:2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683',
            b'prefix:cd42404d52ad55ccfa9aca4adc828aa5800ad9d385a0671fbcbf724118320619',
            b'prefix:2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683',
        ]
        # Create a key with another prefix
        obj = RedisCache(prefix='other')
        obj.flush_all()
        obj.set('key', 1)

        obj = RedisCache(prefix='prefix')
        obj.set('key', 1)
        obj.set('value', 1)
        self.assertListEqual(sorted(obj.connection.keys()), sorted(keys))

        # Delete the key with other
        obj.delete_prefix(prefix='other')
        self.assertListEqual(sorted(obj.connection.keys()), sorted(keys[1:]))

        # Delete keys with prefix
        obj.delete_prefix()
        self.assertListEqual(obj.connection.keys(), [])


class RedisCacheGetSetTestCase(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        RedisClient._connection = None
        with mock.patch.object(redis_module.log, 'debug') as debug:
            cls.cache = RedisCache()
            debug.assert_called_once_with('Connecting on Redis(%s).....', '0.0.0.0')

        cls.key = 'redis-get-set-test-case'

    def setUp(self) -> None:
        self.cache.delete(self.key)
        self.mock = mock.MagicMock()

    def func(self, **kwargs):
        sleep(0.1)
        self.mock(**kwargs)
        return f'return: {kwargs}'

    def test_redis_get_set(self):
        obj01 = RedisCache()
        obj02 = RedisCache()
        with ThreadPoolExecutor() as executor:
            task01 = executor.submit(obj01.get_set, key=self.key, func=self.func, keyword='value01')
            sleep(0.05) # This is just to guarantee that value1 will always be called first
            task02 = executor.submit(obj02.get_set, key=self.key, func=self.func, keyword='value02')
            self.assertEqual(task01.result(), task02.result())
        self.mock.assert_called_once_with(keyword='value01')

    @mock.patch.object(redis_module.log, 'error')
    def test_get_set_func_error_both_exit(self, error: mock.MagicMock):
        self.mock.side_effect = AttributeError
        obj01 = RedisCache()
        obj02 = RedisCache()
        with ThreadPoolExecutor() as executor:
            task01 = executor.submit(obj01.get_set, key=self.key, func=self.func, keyword='value01')
            sleep(0.05) # This is just to guarantee that value1 will always be called first
            task02 = executor.submit(obj02.get_set, key=self.key, func=self.func, keyword='value02')
            self.assertIsNone(task01.result())
            self.assertIsNone(task02.result())

        self.mock.assert_called_once_with(keyword='value01')
        error.assert_called_once()

    @mock.patch.object(redis_module.log, 'error')
    def test_get_set_func_error(self, error: mock.MagicMock):
        self.mock.side_effect = AttributeError
        result = RedisCache().get_set(key=self.key, func=self.func)
        self.assertIsNone(result)
        self.mock.assert_called_once_with()
        error.assert_called_once()


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisCacheCompressedTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None

    def test_get_hash_key(self):
        obj = RedisCacheCompressed()
        hash_key = '2c70e12b7a0646f92279f427c7b38e7334d8e5389cff167a1dc30e73f826b683'
        self.assertEqual(obj.get_hash_key('key'), hash_key)
        # Must always be the same value
        self.assertEqual(obj.get_hash_key('key'), hash_key)

    def test_get_set_delete(self):
        obj = RedisCacheCompressed()
        self.assertIsNone(obj.get('key'))
        lst = [1, 'test', 1.1, {'key': 'value'}]
        obj.set('key', lst)
        self.assertListEqual(obj.get('key'), lst)
        dct = {'lst': [1, 't'], 'dct': {'key': 'value'}, 'key': 1}
        obj.set('key', dct)
        self.assertDictEqual(obj.get('key'), dct)
        obj.delete('key')
        self.assertIsNone(obj.get('key'))


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisListTestCase(TestCase):

    def setUp(self) -> None:
        RedisClient._connection = None
        self.name = 'redis-list-test-case'

    def test_push_pop(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.pop)
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        self.assertEqual(lst.pop(), obj)

    def test_bpop_timeout(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.bpop, 0.5)

    def test_bpop(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        self.assertTupleEqual(lst.bpop(0.5), (self.name, obj))

    def test_clear(self):
        lst = RedisList(name=self.name)
        lst.clear() # Clear all elements
        self.assertRaises(RedisEmptyListError, lst.pop)
        obj = {'lst': [1, 1.1, 'test', True]}
        lst.push(obj)
        lst.clear()
        self.assertRaises(RedisEmptyListError, lst.pop)


@mock.patch.object(redis_module, 'log', mock.MagicMock())
@mock.patch.object(RedisChannel, '_channel', spec=client.PubSub)
class RedisChannelTestCase(TestCase):

    def setUp(self) -> None:
        self.name = 'tests'
        RedisClient._connection = None

    @mock.patch.object(RedisClient, '_connection', spec=redis_module.Redis)
    def test_send(self, _connection: mock.MagicMock, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        channel.send('Hi')
        _connection.publish.assert_called_once_with(channel.name, 'Hi')
        _channel.assert_not_called()

    @mock.patch.object(RedisClient, '_connection', spec=redis_module.Redis)
    def test_channel(self, _connection: mock.MagicMock, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        channel._channel = None
        _connection.pubsub.return_value = _channel
        channel.channel # Call the property - pylint: disable=pointless-statement
        _connection.pubsub.assert_called_once_with()
        _channel.subscribe.assert_called_once_with(channel.name)

    def test_parse_message(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        self.assertTupleEqual(channel.parse_message(message), (self.name, 'data'))
        _channel.assert_not_called()

    def test_consume_process_message(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        _channel.listen.return_value = [message]
        channel.process_message = mock.MagicMock()
        channel.consume()
        channel.process_message.assert_called_once_with('data')

    def test_process(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        self.assertIsNone(channel.process_message('data'))
        _channel.assert_not_called()

    def test_consume_callback(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        _channel.listen.return_value = [message]
        callback = mock.MagicMock()
        channel.consume(callback)
        callback.assert_called_once_with('data')

    def test_consume_break(self, _channel: mock.MagicMock):
        channel = RedisChannel(name=self.name)
        message = {'channel': self.name.encode('utf-8'), 'data': b'data'}
        exit_message = {'channel': self.name.encode('utf-8'), 'data': RedisChannel.exit_message}
        _channel.listen.return_value = [exit_message, message]
        callback = mock.MagicMock()
        channel.consume(callback)
        callback.assert_not_called()


@mock.patch.object(redis_module, 'log', mock.MagicMock())
class RedisLockTestCase(TestCase):

    def setUp(self) -> None:
        self.name = 'tests'
        RedisClient._connection = None

    def tearDown(self) -> None:
        RedisLock(name=self.name).release(force=True)

    def test_acquire(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        self.assertTrue(lock01.acquire())
        self.assertFalse(lock02.acquire())

    @mock.patch.object(RedisLock, '_lock')
    def test_acquire_blocking(self, lock: mock.MagicMock):
        lock01 = RedisLock(name=self.name)
        lock01.acquire(blocking=True, blocking_timeout=1.0)
        lock.acquire.assert_called_once_with(blocking=True, blocking_timeout=1.0)

    def test_timeout(self):
        lock01 = RedisLock(name=self.name, timeout=1)
        lock02 = RedisLock(name=self.name, timeout=1)
        self.assertTrue(lock01.acquire())
        self.assertFalse(lock02.acquire())
        sleep(1.5) # With 1 second the test breaks sometimes
        self.assertTrue(lock02.acquire())

    def test_owned(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire()
        lock02.acquire()
        self.assertTrue(lock01.owned())
        self.assertFalse(lock02.owned())

    def test_release(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire()
        lock02.acquire()
        self.assertFalse(lock02.release())
        self.assertFalse(lock02.acquire())
        self.assertTrue(lock01.release())
        self.assertTrue(lock02.acquire())

    def test_release_force(self):
        lock01 = RedisLock(name=self.name)
        lock02 = RedisLock(name=self.name)
        lock01.acquire()
        lock02.acquire()
        self.assertTrue(lock02.release(force=True))
        self.assertTrue(lock02.acquire())
        self.assertFalse(lock01.release())
        self.assertFalse(lock01.acquire())



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import os
from json import JSONDecodeError
from everysk.core.datetime import Date, DateTime
from everysk.core.object import BaseObject, BaseDict
from everysk.core.serialize import dumps, loads
from everysk.core.unittests import TestCase


PDF = (
    'JVBERi0xLjcKJcOkw7zDtsOfCjIgMCBvYmoKPDwvTGVuZ3RoIDMgMCBSL0ZpbHRlci9GbGF0ZURlY29kZT4+CnN0cmVhbQp4nDPQM1Qo59J3LCrJTEtMLlFw8nXmKlQ'
    'wUDDQMzCyUDC1NNUzMjdVsDAx1LMwM1QoSuUK11LI43IFqgpUAAC+aw3mCmVuZHN0cmVhbQplbmRvYmoKCjMgMCBvYmoKNjcKZW5kb2JqCgo2IDAgb2JqCjw8Cj4+Cm'
    'VuZG9iagoKNyAwIG9iago8PAovRm9udCA2IDAgUgovUHJvY1NldFsvUERGL1RleHRdCj4+CmVuZG9iagoKMSAwIG9iago8PC9UeXBlL1BhZ2UvUGFyZW50IDUgMCBSL'
    '1Jlc291cmNlcyA3IDAgUi9NZWRpYUJveFswIDAgNTk1LjMwMzkzNzAwNzg3NCA4NDEuODg5NzYzNzc5NTI4XS9UYWJzL1MKL0NvbnRlbnRzIDIgMCBSPj4KZW5kb2Jq'
    'Cgo0IDAgb2JqCjw8L1R5cGUvU3RydWN0RWxlbQovUy9Eb2N1bWVudAovUCA4IDAgUgovUGcgMSAwIFIKPj4KZW5kb2JqCgo4IDAgb2JqCjw8L1R5cGUvU3RydWN0VHJ'
    'lZVJvb3QKL1BhcmVudFRyZWUgOSAwIFIKL0tbNCAwIFIgIF0KPj4KZW5kb2JqCgo5IDAgb2JqCjw8L051bXNbCl0+PgplbmRvYmoKCjUgMCBvYmoKPDwvVHlwZS9QYW'
    'dlcwovUmVzb3VyY2VzIDcgMCBSCi9LaWRzWyAxIDAgUiBdCi9Db3VudCAxPj4KZW5kb2JqCgoxMCAwIG9iago8PC9UeXBlL0NhdGFsb2cvUGFnZXMgNSAwIFIKL1BhZ'
    '2VNb2RlL1VzZU91dGxpbmVzCi9PcGVuQWN0aW9uWzEgMCBSIC9YWVogbnVsbCBudWxsIDBdCi9TdHJ1Y3RUcmVlUm9vdCA4IDAgUgovTGFuZyhwdC1CUikKL01hcmtJ'
    'bmZvPDwvTWFya2VkIHRydWU+Pgo+PgplbmRvYmoKCjExIDAgb2JqCjw8L0NyZWF0b3I8RkVGRjAwNTcwMDcyMDA2OTAwNzQwMDY1MDA3Mj4KL1Byb2R1Y2VyPEZFRkY'
    'wMDRDMDA2OTAwNjIwMDcyMDA2NTAwNEYwMDY2MDA2NjAwNjkwMDYzMDA2NTAwMjAwMDMyMDAzNDAwMkUwMDMyPgovQ3JlYXRpb25EYXRlKEQ6MjAyNTA1MjAxNjEzND'
    'MtMDMnMDAnKT4+CmVuZG9iagoKeHJlZgowIDEyCjAwMDAwMDAwMDAgNjU1MzUgZiAKMDAwMDAwMDI1MiAwMDAwMCBuIAowMDAwMDAwMDE5IDAwMDAwIG4gCjAwMDAwM'
    'DAxNTcgMDAwMDAgbiAKMDAwMDAwMDM4MyAwMDAwMCBuIAowMDAwMDAwNTUzIDAwMDAwIG4gCjAwMDAwMDAxNzYgMDAwMDAgbiAKMDAwMDAwMDE5OCAwMDAwMCBuIAow'
    'MDAwMDAwNDUyIDAwMDAwIG4gCjAwMDAwMDA1MjQgMDAwMDAgbiAKMDAwMDAwMDYyNiAwMDAwMCBuIAowMDAwMDAwNzkzIDAwMDAwIG4gCnRyYWlsZXIKPDwvU2l6ZSA'
    'xMi9Sb290IDEwIDAgUgovSW5mbyAxMSAwIFIKL0lEIFsgPDQzMTkzRDFENDQ4NDI4MUFCQ0ZCMzYzMEU0RkMzNDY3Pgo8NDMxOTNEMUQ0NDg0MjgxQUJDRkIzNjMwRT'
    'RGQzM0Njc+IF0KL0RvY0NoZWNrc3VtIC8yMEIxOTM4NTMzNTVGQUQwMjFBRUU5MEU2ODFGMDk2QQo+PgpzdGFydHhyZWYKOTcyCiUlRU9GCg=='
)

PDF_PATH = os.path.join(os.path.dirname(__file__), 'example.pdf')


class MyDict(dict):
    pass

class SerializeOrjsonDumpsTestCase(TestCase):

    def setUp(self):
        self.params: dict = {
            'protocol': 'orjson',
            'sort_keys': True
        }

    ## Constants
    def test_none(self):
        self.assertEqual(dumps(None, **self.params), 'null')

    def test_undefined_true(self):
        self.assertEqual(dumps(Undefined, use_undefined=True, **self.params), '{"__undefined__":null}')

    def test_undefined_false(self):
        self.assertEqual(dumps(Undefined, use_undefined=False, **self.params), 'null')

    def test_nan(self):
        self.assertEqual(dumps(float('NaN'), use_undefined=False, **self.params), 'null')

    def test_infinity(self):
        self.assertEqual(dumps(float('Infinity'), use_undefined=False, **self.params), 'null')

    def test_minus_infinity(self):
        self.assertEqual(dumps(float('-Infinity'), use_undefined=False, **self.params), 'null')

    ## Basic types
    def test_str(self):
        self.assertEqual(dumps('teste', **self.params), '"teste"')

    def test_bytes(self):
        self.assertEqual(dumps({'key': b'teste'}, **self.params), '{"key":{"__bytes__":{"encoding":"utf-8","value":"teste"}}}')

    def test_bytes_base64(self):
        with open(PDF_PATH, 'rb') as fd:
            data = fd.read()

        self.assertEqual(
            dumps({'key': data}, **self.params),
            (
                '{"key":{"__bytes__":{"encoding":"base64","value":"'
                f'{PDF}'
                '"}}}'
            )
        )

    def test_int(self):
        self.assertEqual(dumps(1, **self.params), '1')

    def test_float(self):
        self.assertEqual(dumps(1.0, **self.params), '1.0')

    def test_true(self):
        self.assertEqual(dumps(True, **self.params), 'true')

    def test_false(self):
        self.assertEqual(dumps(False, **self.params), 'false')

    def test_tuple(self):
        self.assertEqual(dumps((2, 'a', True), **self.params), '[2,"a",true]')

    def test_list(self):
        self.assertEqual(dumps([2, 'a', True], **self.params), '[2,"a",true]')

    def test_set(self):
        # set have random order so we use only one element
        self.assertEqual(dumps({2}, **self.params), '[2]')

    def test_dict(self):
        self.assertEqual(dumps({'a': 1, '2': 'b'}, **self.params), '{"2":"b","a":1}')

    def test_date(self):
        self.assertEqual(
            dumps(Date(2025, 1, 1), **self.params),
            '{"__date__":"2025-01-01"}'
        )

    def test_date_with_format(self):
        self.assertEqual(
            dumps(Date(2025, 1, 1), date_format='%Y-%m-%d', **self.params),
            '"2025-01-01"'
        )

    def test_datetime(self):
        self.assertEqual(
            dumps(DateTime(2025, 1, 1, 12, 0, 0), **self.params),
            '{"__datetime__":"2025-01-01T12:00:00+00:00"}'
        )
    def test_datetime_with_format(self):
        self.assertEqual(
            dumps(DateTime(2025, 1, 1, 12, 0, 0), datetime_format='%Y-%m-%d %H:%M:%S', **self.params),
            '"2025-01-01 12:00:00"'
        )

    ## Objects
    def test_base_object(self):
        self.assertEqual(
            dumps(BaseObject(a=1, b=2), **self.params),
            '{"__class_path__":"everysk.core.object.BaseObject","_errors":null,"_is_frozen":false,"_silent":false,"a":1,"b":2}'
        )

    def test_base_object_without_class_path(self):
        self.assertEqual(
            dumps(BaseObject(a=1, b=2), add_class_path=False, **self.params),
            '{"a":1,"b":2}'
        )

    def test_base_dict(self):
        self.assertEqual(
            dumps(BaseDict(a=1, b=2), **self.params),
            '{"__class_path__":"everysk.core.object.BaseDict","_errors":null,"_is_frozen":false,"_silent":false,"a":1,"b":2}'
        )

    def test_base_dict_without_class_path(self):
        self.assertEqual(
            dumps(BaseDict(a=1, b=2), add_class_path=False, **self.params),
            '{"a":1,"b":2}'
        )

    def test_dict_subclass_default(self):
        # This use default dict serializer
        self.assertEqual(
            dumps(MyDict(a=1, b=2), **self.params),
            '{"a":1,"b":2}'
        )

    def test_dict_subclass_passthrough(self):
        # This will raise a TypeError because MyDict is not implemented
        # to be serializable inside the JSONEncoder class
        self.assertRaises(
            TypeError,
            dumps,
            MyDict(a=1, b=2),
            passthrough_subclass=True,
            **self.params
        )

    def test_dict_key_non_str(self):
        self.assertEqual(
            dumps({1: 'a', 1.1: 'b'}),
            '{"1": "a", "1.1": "b"}'
        )

    def test_return_str(self):
        self.assertEqual(dumps('string', protocol='orjson', return_type='str'), '"string"')

    def test_return_bytes(self):
        self.assertEqual(dumps('string', protocol='orjson', return_type='bytes'), b'"string"')


class SerializeOrjsonLoadsTestCase(TestCase):

    def setUp(self):
        self.params: dict = {
            'protocol': 'orjson'
        }

    ## Constants
    def test_null(self):
        self.assertEqual(loads(b'null', **self.params), None)

    def test_undefined_true(self):
        self.assertEqual(loads(b'{"__undefined__":null}', use_undefined=True, **self.params), Undefined)

    def test_undefined_false(self):
        self.assertEqual(loads(b'{"__undefined__":null}', use_undefined=False, **self.params), None)

    def test_nan(self):
        self.assertRaises(JSONDecodeError, loads, 'NaN', **self.params)

    def test_infinity(self):
        self.assertRaises(JSONDecodeError, loads, 'Infinity', **self.params)

    def test_minus_infinity(self):
        self.assertRaises(JSONDecodeError, loads, '-Infinity', **self.params)

    ## Basic types
    def test_str(self):
        self.assertEqual(loads(b'"teste"', **self.params), 'teste')

    def test_bytes(self):
        self.assertEqual(loads(b'{"__bytes__":{"encoding":"utf-8","value":"teste"}}', **self.params), b'teste')

    def test_bytes_base64(self):
        with open(PDF_PATH, 'rb') as fd:
            data = fd.read()

        self.assertEqual(
            loads(
                (
                    '{"key":{"__bytes__":{"encoding":"base64","value":"'
                    f'{PDF}'
                    '"}}}'
                ).encode('utf-8'),
                **self.params
            ),
            {'key': data}
        )

    def test_int(self):
        self.assertEqual(loads(b'1', **self.params), 1)

    def test_float(self):
        self.assertEqual(loads(b'1.0', **self.params), 1.0)

    def test_true(self):
        self.assertEqual(loads(b'true', **self.params), True)

    def test_false(self):
        self.assertEqual(loads(b'false', **self.params), False)

    def test_list(self):
        self.assertEqual(loads(b'[2,"a",true]', **self.params), [2, 'a', True])

    def test_dict(self):
        self.assertEqual(loads(b'{"a":1,"2":"b"}', **self.params), {'a': 1, '2': 'b'})

    def test_date(self):
        self.assertEqual(
            loads(b'{"__date__":"2025-01-01"}', **self.params),
            Date(2025, 1, 1)
        )

    def test_datetime(self):
        self.assertEqual(
            loads(b'{"__datetime__":"2025-01-01T12:00:00+00:00"}', **self.params),
            DateTime(2025, 1, 1, 12, 0, 0)
        )


    ## Objects
    def test_base_object(self):
        # pylint: disable=protected-access
        obj = loads(
            b'{"__class_path__":"everysk.core.object.BaseObject","_errors":null,"_is_frozen":false,"_silent":false,"a":1,"b":2}',
            **self.params
        )
        self.assertEqual(obj.a, 1)
        self.assertEqual(obj.b, 2)
        self.assertEqual(obj._errors, None)
        self.assertEqual(obj._is_frozen, False)
        self.assertEqual(obj._silent, False)
        self.assertEqual(obj.__class__.__name__, 'BaseObject')

    def test_base_dict(self):
        self.assertEqual(
            loads(
                b'{"__class_path__":"everysk.core.object.BaseDict","_errors":null,"_is_frozen":false,"_silent":false,"a":1,"b":2}',
                **self.params
            ),
            BaseDict(a=1, b=2)
        )

    def test_dict_subclass(self):
        self.assertEqual(
            loads(
                b'{"__class_path__":"everysk.core._tests.serialize.test_orjson.MyDict","a":1,"b":2}',
                **self.params
            ),
            MyDict(a=1, b=2)
        )



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.core.datetime import DateTime
from everysk.core.serialize import dumps, loads, CLASS_KEY
from everysk.core.unittests import TestCase, mock
from everysk.sdk.worker_base import WorkerBase
from everysk.sdk.engines.cache import UserCache
from everysk.sdk.engines.market_data import MarketData
from everysk.sdk.engines.expression import Expression
from everysk.sdk.engines.compliance import Compliance
from everysk.sdk.entities.portfolio.base import Portfolio
from everysk.sdk.entities.portfolio.security import Security
from everysk.sdk.entities.portfolio.securities import Securities
from everysk.sdk.entities.custom_index.base import CustomIndex
from everysk.sdk.entities.datastore.base import Datastore
from everysk.sdk.entities.file.base import File
from everysk.sdk.entities.private_security.base import PrivateSecurity
from everysk.sdk.entities.query import Query
from everysk.sdk.entities.report.base import Report
from everysk.sdk.entities.script import Script


@mock.patch.object(DateTime, 'now', mock.MagicMock(return_value=DateTime(2024, 6, 20, 14, 38, 59, 360554)))
class SerializeDumpsSDKTestCase(TestCase):
    maxDiff = None

    def test_sdk_objects_worker_base(self):
        self.assertEqual(
            dumps(WorkerBase({}), sort_keys=True),
            '{"%s": "everysk.sdk.worker_base.WorkerBase", "_errors": null, "_is_frozen": false, "_silent": false, "inputs_info": null, "parallel_info": null, "script_inputs": null, "worker_id": null, "worker_type": null, "workflow_execution_id": null, "workflow_id": null, "workspace": null}' % CLASS_KEY
        )

    def test_sdk_engines_objects_user_cache(self):
        self.assertEqual(
            dumps(UserCache(), sort_keys=True),
            '{"%s": "everysk.sdk.engines.cache.UserCache", "_errors": null, "_is_frozen": false, "_silent": false}' % CLASS_KEY
        )

    def test_sdk_engines_objects_market_data(self):
        self.assertEqual(
            dumps(MarketData(), sort_keys=True),
            '{"%s": "everysk.sdk.engines.market_data.MarketData", "_errors": null, "_is_frozen": false, "_silent": false, "date": null, "end_date": null, "nearest": false, "projection": null, "real_time": false, "start_date": null, "ticker_list": null, "ticker_type": null}' % CLASS_KEY
        )

    def test_sdk_engines_objects_expression(self):
        self.assertEqual(
            dumps(Expression(), sort_keys=True),
            '{"%s": "everysk.sdk.engines.expression.Expression", "_errors": null, "_is_frozen": false, "_silent": false}' % CLASS_KEY
        )

    def test_sdk_engines_objects_compliance(self):
        self.assertEqual(
            dumps(Compliance(), sort_keys=True),
            '{"%s": "everysk.sdk.engines.compliance.Compliance", "_errors": null, "_is_frozen": false, "_silent": false}' % CLASS_KEY
        )

    def test_sdk_entities_objects_portfolio(self):
        self.assertEqual(
            dumps(Portfolio(), sort_keys=True),
            '{"%s": "everysk.sdk.entities.portfolio.base.Portfolio", "base_currency": null, "check_securities": false, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "nlv": null, "securities": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}' % CLASS_KEY
        )

    def test_sdk_entities_objects_security(self):
        self.assertEqual(
            dumps(Security(), sort_keys=True),
            '{"%s": "everysk.sdk.entities.portfolio.security.Security", "accounting": null, "asset_class": null, "asset_subclass": null, "book": null, "comparable": null, "cost_price": null, "coupon": null, '
            '"currency": null, "display": null, "error_message": null, "error_type": null, "exchange": null, "extra_data": null, "fx_rate": null, "hash": null, "id": null, "indexer": null, "instrument_class": null, "instrument_subtype": null, "instrument_type": null, "isin": null, "issue_date": null, '
            '"issue_price": null, "issuer": null, "issuer_type": null, "label": null, "look_through_reference": null, "market_price": null, "market_value": null, "market_value_in_base": null, "maturity_date": null, "multiplier": null, "name": null, "operation": null, "option_type": null, "percent_index": null, '
            '"premium": null, "previous_quantity": null, "quantity": null, "rate": null, "return_date": null, "series": null, "settlement": null, "status": null, "strike": null, "symbol": null, "ticker": null, "trade_id": null, "trader": null, "underlying": null, "unrealized_pl": null, "unrealized_pl_in_base": null, "warranty": null}' % CLASS_KEY
        )

    def test_sdk_entities_objects_securities(self):
        self.assertEqual(dumps(Securities(), sort_keys=True), '[]')

    def test_sdk_entities_objects_custom_index(self):
        self.assertEqual(
            dumps(CustomIndex(), sort_keys=True),
            '{"%s": "everysk.sdk.entities.custom_index.base.CustomIndex", "base_price": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "data_type": null, "description": null, "id": null, "name": null, "periodicity": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}' % CLASS_KEY
        )

    def test_sdk_entities_objects_datastore(self):
        self.assertEqual(
            dumps(Datastore(), sort_keys=True),
            '{"%s": "everysk.sdk.entities.datastore.base.Datastore", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}' % CLASS_KEY
        )

    def test_sdk_entities_objects_file(self):
        self.assertEqual(
            dumps(File(), sort_keys=True),
            '{"%s": "everysk.sdk.entities.file.base.File", "content_type": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "hash": null, "id": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "workspace": null}' % CLASS_KEY
        )

    def test_sdk_entities_objects_private(self):
        self.assertEqual(
            dumps(PrivateSecurity(), sort_keys=True),
            '{"%s": "everysk.sdk.entities.private_security.base.PrivateSecurity", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "description": null, "id": null, "instrument_type": null, "name": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}' % CLASS_KEY
        )

    def test_sdk_entities_objects_query(self):
        self.assertEqual(
            dumps(Query(Portfolio), sort_keys=True),
            '{"%s": "everysk.sdk.entities.query.Query", "_clean_order": [], "_find_or_fail": false, "_klass": "Portfolio", "distinct_on": [], "filters": [], "limit": null, "offset": null, "order": [], "page_size": null, "page_token": null, "projection": null}' % CLASS_KEY
        )

    def test_sdk_entities_objects_report(self):
        self.assertEqual(
            dumps(Report(), sort_keys=True),
            '{"%s": "everysk.sdk.entities.report.base.Report", "authorization": null, "config_cascaded": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "layout_content": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "widgets": null, "workspace": null}' % CLASS_KEY
        )

    def test_sdk_entities_objects_script(self):
        self.assertEqual(
            dumps(Script(Portfolio), sort_keys=True),
            '{"%s": "everysk.sdk.entities.script.Script", "_klass": "Portfolio"}' % CLASS_KEY
        )


@mock.patch.object(DateTime, 'now', mock.MagicMock(return_value=DateTime(2024, 6, 20, 14, 38, 59, 360554)))
class SerializeLoadsSDKTestCase(TestCase):

    def test_sdk_objects_worker_base(self):
        self.assertEqual(
            loads('{"%s": "everysk.sdk.app.worker_base.WorkerBase", "inputs_info": null, "parallel_info": null, "script_inputs": null, "worker_id": null, "worker_type": null, "workflow_execution_id": null, "workflow_id": null, "workspace": null}' % CLASS_KEY),
            WorkerBase({})
        )

    def test_sdk_engines_objects_user_cache(self):
        self.assertIsInstance(loads('{"%s": "everysk.sdk.app.engines.cache.UserCache"}' % CLASS_KEY), UserCache)

    def test_sdk_engines_objects_security(self):
        self.assertIsInstance(loads('{"%s": "everysk.sdk.app.engines.market_data.MarketData"}' % CLASS_KEY), MarketData)

    def test_sdk_engines_objects_market_data(self):
        self.assertIsInstance(loads('{"%s": "everysk.sdk.app.engines.expression.Expression"}' % CLASS_KEY), Expression)

    def test_sdk_engines_objects_compliance(self):
        self.assertIsInstance(loads('{"%s": "everysk.sdk.app.engines.compliance.Compliance"}' % CLASS_KEY), Compliance)

    def test_sdk_entities_objects_portfolio(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.portfolio.base.Portfolio", "base_currency": null, "check_securities": false, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "nlv": null, "securities": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}' % CLASS_KEY),
            Portfolio()
        )

    def test_sdk_entities_objects_security(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.portfolio.security.Security", "accounting": null, "asset_class": null, "asset_subclass": null, "book": null, "comparable": null, "cost_price": null, "coupon": null, "currency": null, "display": null, "error_message": null, "error_type": null, "exchange": null, "extra_data": null, "fx_rate": null, "hash": null, "id": null, "indexer": null, "instrument_class": null, "instrument_subtype": null, "instrument_type": null, "isin": null, "issue_date": null, "issue_price": null, "issuer": null, "issuer_type": null, "label": null, "look_through_reference": null, "market_price": null, "market_value": null, "market_value_in_base": null, "maturity_date": null, "multiplier": null, "name": null, "operation": null, "option_type": null, "percent_index": null, "premium": null, "previous_quantity": null, "quantity": null, "rate": null, "return_date": null, "series": null, "settlement": null, "status": null, "strike": null, "symbol": null, "ticker": null, "trade_id": null, "trader": null, "underlying": null, "unrealized_pl": null, "unrealized_pl_in_base": null, "warranty": null}' % CLASS_KEY),
            Security()
        )

    def test_sdk_entities_objects_securities(self):
        self.assertEqual(loads('[]'), Securities())

    def test_sdk_entities_objects_custom_index(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.custom_index.base.CustomIndex", "base_price": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "data_type": null, "description": null, "name": null, "periodicity": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}' % CLASS_KEY),
            CustomIndex()
        )

    def test_sdk_entities_objects_datastore(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.datastore.base.Datastore", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "id": null, "level": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1", "workspace": null}' % CLASS_KEY),
            Datastore()
        )

    def test_sdk_entities_objects_file(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.file.base.File", "content_type": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "data": null, "date": null, "description": null, "hash": null, "id": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "workspace": null}' % CLASS_KEY),
            File()
        )

    def test_sdk_entities_objects_loads(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.private_security.base.PrivateSecurity", "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "currency": null, "data": null, "description": null, "instrument_type": null, "name": null, "symbol": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "version": "v1"}' % CLASS_KEY),
            PrivateSecurity()
        )

    def test_sdk_entities_objects_query(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.query.Query", "_clean_order": [], "_klass": "Portfolio", "distinct_on": [], "filters": [], "limit": null, "offset": null, "order": [], "page_size": null, "page_token": null, "projection": null}' % CLASS_KEY),
            Query(Portfolio)
        )

    def test_sdk_entities_objects_report(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.report.base.Report", "authorization": null, "config_cascaded": null, "created_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "date": null, "description": null, "id": null, "layout_content": null, "link_uid": null, "name": null, "tags": [], "updated_on": {"__datetime__": "2024-06-20T14:38:59.360554+00:00"}, "url": null, "version": "v1", "widgets": null, "workspace": null}' % CLASS_KEY),
            Report()
        )

    def test_sdk_entities_objects_script(self):
        self.assertDictEqual(
            loads('{"%s": "everysk.sdk.app.entities.script.Script", "_klass": "Portfolio"}' % CLASS_KEY),
            Script(Portfolio)
        )

    def test_sdk_entities_when_instantiate_object_is_false(self):
        p = Portfolio()
        result = dumps(p)
        res_loads = loads(result, instantiate_object=False)

        self.assertIsInstance(res_loads, dict)
        self.assertDictEqual(res_loads, Portfolio())



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
import json
import time

from everysk.api import api_requestor, get_api_config


###############################################################################
#   Private Functions Implementation
###############################################################################
def dumps_json(obj):
    """
    Serializes a Python object into a JSON-formatted string with sorted keys and pretty-printing.

    Args:
        obj:
            The Python object to serialize

    Returns:
        str:  A JSON-formatted string representing the serialized output. The keys will be sorted with an level 2 of indentation

    Example:
        >>> from everysk.api.utils import dumps_json
        >>> data = {
        ...     2: 'apple',
        ...     1: 'banana',
        ...     3: 'cherry'
        ... }
        >>> json_str = dumps_json(data)
        >>> print(json_str)
        {
        "1": "banana",
        "2": "apple",
        "3": "cherry"
        }
    """
    return json.dumps(obj, sort_keys=True, indent=2)


###############################################################################
#   Everysk Object Implementation
###############################################################################
class EveryskObject(dict):
    def __init__(self, retrieve_params, params) -> None: # pylint: disable=unused-argument
        super().__init__()
        self.__unsaved_values = set()
        self.update(params)
        self.clear_unsaved_values()

    def get_unsaved_values(self):
        return {k:self[k] for k in self.__unsaved_values}

    def clear_unsaved_values(self):
        self.__unsaved_values = set()

    def __str__(self):
        return dumps_json(self)

    def __repr__(self):
        ident_parts = [type(self).__name__]
        ident_parts.append(f"id={self.get('id')}")
        unicode_repr = f"<{' '.join(ident_parts)} at {hex(id(self))}> JSON: {str(self)}"
        return unicode_repr

    def update(self, dict_) -> None:
        for k, v in dict_.items():
            self.__setattr__(k, v)

    def __setattr__(self, k, v):
        if k.startswith('_') or k in self.__dict__:
            return super().__setattr__(k, v)

        self[k] = v
        return None

    def __getattr__(self, k):
        if k.startswith('_'):
            raise AttributeError(k)

        try:
            return self[k]
        except KeyError as err:
            raise AttributeError(*err.args) # pylint: disable=raise-missing-from

    def __delattr__(self, k):
        if k.startswith('_') or k in self.__dict__:
            return super().__delattr__(k)
        else:
            del self[k]

    def __setitem__(self, k, v) -> None:
        self.__unsaved_values.add(k)
        super().__setitem__(k, v)

    def __getitem__(self, k):
        return super().__getitem__(k)

    def __delitem__(self, k) -> None:
        super().__delitem__(k)
        self.__unsaved_values.remove(k)

        # # Allows for unpickling in Python 3.x
        # if hasattr(self, '_unsaved_values'):
        #     self._unsaved_values.remove(k)


###############################################################################
#   Everysk List Implementation
###############################################################################
class EveryskList(list):
    def __init__(self, retrieve_params, response, key, cls) -> None:
        """
        Initializes the EveryskList object.

        Args:
            retrieve_params (dict): The parameters used for retrieving data.
            response (dict): The response containing the data.
            key (str): The key in the response dictionary that contains the data.
            cls (class): The class used to create objects from the data.
        """
        super().__init__()
        self.__page_size = retrieve_params.get('page_size', 10)
        self.__next_page_token = response.get('next_page_token', None)
        self.extend([cls({}, params) for params in response[key]])

    def page_size(self):
        """
        Returns the page size used for pagination.

        Returns:
            int: The page size used for pagination.
        """
        return self.__page_size

    def next_page_token(self):
        """
        Retrieves the next page token.

        Returns:
            The next page token.
        """
        return self.__next_page_token


###############################################################################
#   Private Functions Implementation
###############################################################################
def create_api_requestor(params=None):
    """
    Creates an APIRequestor object configured with API settings extracted or derived from the provided parameters.
    It initializes an APIRequestor with configuration settings such as API token, API session ID, API entry point, and SSL certificates status.

    Args:
        params (dict, optional):
            A dictionary of parameters containing API configuration settings. Defaults to an empty dictionary.

    Returns:
        APIRequestor: An instance of APIRequestor, configured with the API settings

    Example:
        >>> api_config_params = {
        >>> ... 'api_entry': 'https://api.example.com',
        >>> ... 'api_version': 'v1',
        >>> ... 'api_sid': 'session123',
        >>> ... 'api_token': 'token456',
        >>> ... 'verify_ssl_certs': False
        >>> }
        >>> api_requestor = create_api_requestor(api_config_params)
        >>> print(api_requestor.base_url)  # This will output the base URL configured for the API requestor.
    """
    params = params or {}
    return api_requestor.APIRequestor(
        *get_api_config(params)
    )

def to_object(cls, retrieve_params, response, extra_keys=None):
    """
    Converts a response dictionary into an instance of the specified class.

    Args:
        cls (type): The class to instantiate. It must implement a constructor
            that accepts (retrieve_params, response_body), and a `class_name()` method.
        retrieve_params (dict): Parameters originally used to retrieve the data.
        response (dict): The response dictionary received from an API call.
        extra_keys (list[str], optional): List of keys to extract from the top-level
            of the response and merge into the response body. Defaults to None.

    Returns:
        object or None: An instance of `cls` populated with the response data,
        or None if the response is invalid or missing the expected key.

    Example:
        >>> from everysk.api.utils import to_object
        >>> from everysk.api.api_resources import Calculation

        >>> response = {'calculation': {'some': 'data'}, 'metadata': {'id': 123}}
        >>> retrieve_params = {'param1': 'value1'}

        >>> calculation_object = to_object(Calculation, retrieve_params, response, extra_keys=['metadata'])
        # Creates a Calculation instance with data: {'some': 'data', 'metadata': {'id': 123}}
    """
    key = cls.class_name()
    result = None
    if response and key in response and isinstance(response[key], dict):
        response_body = response[key]

        if extra_keys and isinstance(extra_keys, list):
            for key in extra_keys:
                if key in response:
                    response_body[key] = response[key]

        result = cls(retrieve_params, response_body)
    return result

def to_list(cls, retrieve_params, response):
    """
    Convert a response dictionary into a list of objects of the given class.

    Args:
        cls (class): The class of the objects to be created.
        retrieve_params (dict): The retrieve parameters used to fetch the response.
        response (dict): The response dictionary.

    Returns:
        list: A list of objects of the given class, created from the response dictionary.
              If the response is empty, the key is missing, or the value is not a list,
              None is returned.

    Example:
        >>> from unittest.mock import MagicMock
        >>> from everysk.api.utils import to_list
        >>> mock_class = MagicMock()

        >>> mock_class.class_name_list.return_value = 'valid_key'

        >>> response = {'valid_key': [{'some': 'data'}, {'more': 'data'}]}
        >>> retrieve_params = {'param1': 'value1'}

        >>> result = to_list(mock_class, retrieve_params, response)
    """
    key = cls.class_name_list()
    result = None
    if response and key in response and isinstance(response[key], list):
        result = EveryskList(retrieve_params, response, key, cls)
    return result

def sleep(t) -> None:
    """
    Sleeps for the specified number of seconds.

    Args:
        t (float): The number of seconds to sleep.
    """
    time.sleep(t)



###############################################################################
#
# (C) Copyright 2025 EVERYSK TECHNOLOGIES
#
# This is an unpublished work containing confidential and proprietary
# information of EVERYSK TECHNOLOGIES. Disclosure, use, or reproduction
# without authorization of EVERYSK TECHNOLOGIES is prohibited.
#
###############################################################################
from everysk.api.api_resources.workspace import Workspace
from everysk.api.api_resources.workflow import Workflow
from everysk.api.api_resources.workflow_execution import WorkflowExecution
from everysk.api.api_resources.worker_execution import WorkerExecution
from everysk.api.api_resources.portfolio import Portfolio
from everysk.api.api_resources.report import Report
from everysk.api.api_resources.report_template import ReportTemplate
from everysk.api.api_resources.datastore import Datastore
from everysk.api.api_resources.file import File
from everysk.api.api_resources.custom_index import CustomIndex
from everysk.api.api_resources.private_security import PrivateSecurity
from everysk.api.api_resources.calculation import Calculation
from everysk.api.api_resources.market_data import MarketData
from everysk.api.api_resources.parser import Parser
